import base64
import hashlib
from Interfaces.IKeyPair import IKeyPair
import base58
from Types import Address, Ed25519Signature
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519
import os

class PhantasmaKeys:
    PrivateKeyLength = 32
    PublicKey = None

    def __init__(self, private_key: bytes):
        if len(private_key) == 64:
            private_key = private_key[:32]

        if len(private_key) != self.PrivateKeyLength:
            raise ValueError(f"PrivateKey should have length {self.PrivateKeyLength}")

        self._private_key = private_key
        self._private_key_obj = ed25519.Ed25519PrivateKey.from_private_bytes(self._private_key)
        self._public_key = self._private_key_obj.public_key().public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw
        )
        
        PublicKey = self._public_key
        #if ( hasattr(Address, 'FromKey') ):
        self.Address = Address.FromKey(self._public_key)
        #Address.FromMyKey()  # Implement Address class

    def __str__(self):
        return self.Address.Text

    @staticmethod
    def generate():
        private_key = os.urandom(PhantasmaKeys.PrivateKeyLength)
        return PhantasmaKeys(private_key)

    @staticmethod
    def from_wif(wif: str):

        if not wif:
            raise ValueError("WIF required")

        # Decode the WIF string
        data = base58.b58decode(wif)

        # Perform checks on the decoded data
        if len(data) == 38:
            data = data[:-4]  # Remove the last 4 bytes (checksum)

        if len(data) != 34 or data[0] != 0x80 or data[-1] != 0x01:
            raise ValueError("Invalid WIF format")

        private_key = data[1:33]  # Extract the private key
        return PhantasmaKeys(private_key)

    def to_wif(self):
        # Step 1: Prepend 0x80 to the private key
        extended_private_key = b'\x80' + self._private_key

        # Step 2: Append 0x01 as a suffix
        extended_private_key = extended_private_key + b'\x01'

        # Step 3: Perform SHA-256 hash twice
        first_sha256 = hashlib.sha256(extended_private_key).digest()
        second_sha256 = hashlib.sha256(first_sha256).digest()

        # Step 4: Append the first 4 bytes of the double SHA-256 hash as a checksum
        checksum = second_sha256[:4]
        wif_bytes = extended_private_key + checksum

        # Step 5: Encode the result using base58 encoding
        wif_string = base58.b58encode(wif_bytes)

        return wif_string.decode('utf-8')

    def sign(self, message: bytes):
    # Return an Ed25519Signature object
        return Ed25519Signature.generate(self, message)


# coding: utf-8

"""
    Phantasma API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class Address(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'kind': 'AddressKind',
        'is_system': 'bool',
        'is_interop': 'bool',
        'is_user': 'bool',
        'tendermint_address': 'str',
        'tendermint_public_key': 'str',
        'is_null': 'bool',
        'text': 'str',
        'platform_id': 'int'
    }

    attribute_map = {
        'kind': 'kind',
        'is_system': 'isSystem',
        'is_interop': 'isInterop',
        'is_user': 'isUser',
        'tendermint_address': 'tendermintAddress',
        'tendermint_public_key': 'tendermintPublicKey',
        'is_null': 'isNull',
        'text': 'text',
        'platform_id': 'platformID'
    }

    def __init__(self, kind=None, is_system=None, is_interop=None, is_user=None, tendermint_address=None, tendermint_public_key=None, is_null=None, text=None, platform_id=None):  # noqa: E501
        """Address - a model defined in Swagger"""  # noqa: E501
        self._kind = None
        self._is_system = None
        self._is_interop = None
        self._is_user = None
        self._tendermint_address = None
        self._tendermint_public_key = None
        self._is_null = None
        self._text = None
        self._platform_id = None
        self.discriminator = None
        if kind is not None:
            self.kind = kind
        if is_system is not None:
            self.is_system = is_system
        if is_interop is not None:
            self.is_interop = is_interop
        if is_user is not None:
            self.is_user = is_user
        if tendermint_address is not None:
            self.tendermint_address = tendermint_address
        if tendermint_public_key is not None:
            self.tendermint_public_key = tendermint_public_key
        if is_null is not None:
            self.is_null = is_null
        if text is not None:
            self.text = text
        if platform_id is not None:
            self.platform_id = platform_id

    @property
    def kind(self):
        """Gets the kind of this Address.  # noqa: E501


        :return: The kind of this Address.  # noqa: E501
        :rtype: AddressKind
        """
        return self._kind

    @kind.setter
    def kind(self, kind):
        """Sets the kind of this Address.


        :param kind: The kind of this Address.  # noqa: E501
        :type: AddressKind
        """

        self._kind = kind

    @property
    def is_system(self):
        """Gets the is_system of this Address.  # noqa: E501


        :return: The is_system of this Address.  # noqa: E501
        :rtype: bool
        """
        return self._is_system

    @is_system.setter
    def is_system(self, is_system):
        """Sets the is_system of this Address.


        :param is_system: The is_system of this Address.  # noqa: E501
        :type: bool
        """

        self._is_system = is_system

    @property
    def is_interop(self):
        """Gets the is_interop of this Address.  # noqa: E501


        :return: The is_interop of this Address.  # noqa: E501
        :rtype: bool
        """
        return self._is_interop

    @is_interop.setter
    def is_interop(self, is_interop):
        """Sets the is_interop of this Address.


        :param is_interop: The is_interop of this Address.  # noqa: E501
        :type: bool
        """

        self._is_interop = is_interop

    @property
    def is_user(self):
        """Gets the is_user of this Address.  # noqa: E501


        :return: The is_user of this Address.  # noqa: E501
        :rtype: bool
        """
        return self._is_user

    @is_user.setter
    def is_user(self, is_user):
        """Sets the is_user of this Address.


        :param is_user: The is_user of this Address.  # noqa: E501
        :type: bool
        """

        self._is_user = is_user

    @property
    def tendermint_address(self):
        """Gets the tendermint_address of this Address.  # noqa: E501


        :return: The tendermint_address of this Address.  # noqa: E501
        :rtype: str
        """
        return self._tendermint_address

    @tendermint_address.setter
    def tendermint_address(self, tendermint_address):
        """Sets the tendermint_address of this Address.


        :param tendermint_address: The tendermint_address of this Address.  # noqa: E501
        :type: str
        """

        self._tendermint_address = tendermint_address

    @property
    def tendermint_public_key(self):
        """Gets the tendermint_public_key of this Address.  # noqa: E501


        :return: The tendermint_public_key of this Address.  # noqa: E501
        :rtype: str
        """
        return self._tendermint_public_key

    @tendermint_public_key.setter
    def tendermint_public_key(self, tendermint_public_key):
        """Sets the tendermint_public_key of this Address.


        :param tendermint_public_key: The tendermint_public_key of this Address.  # noqa: E501
        :type: str
        """

        self._tendermint_public_key = tendermint_public_key

    @property
    def is_null(self):
        """Gets the is_null of this Address.  # noqa: E501


        :return: The is_null of this Address.  # noqa: E501
        :rtype: bool
        """
        return self._is_null

    @is_null.setter
    def is_null(self, is_null):
        """Sets the is_null of this Address.


        :param is_null: The is_null of this Address.  # noqa: E501
        :type: bool
        """

        self._is_null = is_null

    @property
    def text(self):
        """Gets the text of this Address.  # noqa: E501


        :return: The text of this Address.  # noqa: E501
        :rtype: str
        """
        return self._text

    @text.setter
    def text(self, text):
        """Sets the text of this Address.


        :param text: The text of this Address.  # noqa: E501
        :type: str
        """

        self._text = text

    @property
    def platform_id(self):
        """Gets the platform_id of this Address.  # noqa: E501


        :return: The platform_id of this Address.  # noqa: E501
        :rtype: int
        """
        return self._platform_id

    @platform_id.setter
    def platform_id(self, platform_id):
        """Sets the platform_id of this Address.


        :param platform_id: The platform_id of this Address.  # noqa: E501
        :type: int
        """

        self._platform_id = platform_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Address, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Address):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# __init__.py
from .Address import Address
from .DomainSettings import DomainSettings
from .Ed25519Signature import Ed25519Signature
from .PhantasmaKeys import PhantasmaKeys
from .Serialization import Serialization
from .Timestamp import Timestamp
from .Stack import Stack

from .Enums import *
from .Extensions import *


# coding: utf-8

"""
    Phantasma API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class ArchiveResult(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'name': 'str',
        'hash': 'str',
        'time': 'int',
        'size': 'int',
        'encryption': 'str',
        'block_count': 'int',
        'missing_blocks': 'list[int]',
        'owners': 'list[str]'
    }

    attribute_map = {
        'name': 'name',
        'hash': 'hash',
        'time': 'time',
        'size': 'size',
        'encryption': 'encryption',
        'block_count': 'blockCount',
        'missing_blocks': 'missingBlocks',
        'owners': 'owners'
    }

    def __init__(self, name=None, hash=None, time=None, size=None, encryption=None, block_count=None, missing_blocks=None, owners=None):  # noqa: E501
        """ArchiveResult - a model defined in Swagger"""  # noqa: E501
        self._name = None
        self._hash = None
        self._time = None
        self._size = None
        self._encryption = None
        self._block_count = None
        self._missing_blocks = None
        self._owners = None
        self.discriminator = None
        if name is not None:
            self.name = name
        if hash is not None:
            self.hash = hash
        if time is not None:
            self.time = time
        if size is not None:
            self.size = size
        if encryption is not None:
            self.encryption = encryption
        if block_count is not None:
            self.block_count = block_count
        if missing_blocks is not None:
            self.missing_blocks = missing_blocks
        if owners is not None:
            self.owners = owners

    @property
    def name(self):
        """Gets the name of this ArchiveResult.  # noqa: E501


        :return: The name of this ArchiveResult.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ArchiveResult.


        :param name: The name of this ArchiveResult.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def hash(self):
        """Gets the hash of this ArchiveResult.  # noqa: E501


        :return: The hash of this ArchiveResult.  # noqa: E501
        :rtype: str
        """
        return self._hash

    @hash.setter
    def hash(self, hash):
        """Sets the hash of this ArchiveResult.


        :param hash: The hash of this ArchiveResult.  # noqa: E501
        :type: str
        """

        self._hash = hash

    @property
    def time(self):
        """Gets the time of this ArchiveResult.  # noqa: E501


        :return: The time of this ArchiveResult.  # noqa: E501
        :rtype: int
        """
        return self._time

    @time.setter
    def time(self, time):
        """Sets the time of this ArchiveResult.


        :param time: The time of this ArchiveResult.  # noqa: E501
        :type: int
        """

        self._time = time

    @property
    def size(self):
        """Gets the size of this ArchiveResult.  # noqa: E501


        :return: The size of this ArchiveResult.  # noqa: E501
        :rtype: int
        """
        return self._size

    @size.setter
    def size(self, size):
        """Sets the size of this ArchiveResult.


        :param size: The size of this ArchiveResult.  # noqa: E501
        :type: int
        """

        self._size = size

    @property
    def encryption(self):
        """Gets the encryption of this ArchiveResult.  # noqa: E501


        :return: The encryption of this ArchiveResult.  # noqa: E501
        :rtype: str
        """
        return self._encryption

    @encryption.setter
    def encryption(self, encryption):
        """Sets the encryption of this ArchiveResult.


        :param encryption: The encryption of this ArchiveResult.  # noqa: E501
        :type: str
        """

        self._encryption = encryption

    @property
    def block_count(self):
        """Gets the block_count of this ArchiveResult.  # noqa: E501


        :return: The block_count of this ArchiveResult.  # noqa: E501
        :rtype: int
        """
        return self._block_count

    @block_count.setter
    def block_count(self, block_count):
        """Sets the block_count of this ArchiveResult.


        :param block_count: The block_count of this ArchiveResult.  # noqa: E501
        :type: int
        """

        self._block_count = block_count

    @property
    def missing_blocks(self):
        """Gets the missing_blocks of this ArchiveResult.  # noqa: E501


        :return: The missing_blocks of this ArchiveResult.  # noqa: E501
        :rtype: list[int]
        """
        return self._missing_blocks

    @missing_blocks.setter
    def missing_blocks(self, missing_blocks):
        """Sets the missing_blocks of this ArchiveResult.


        :param missing_blocks: The missing_blocks of this ArchiveResult.  # noqa: E501
        :type: list[int]
        """

        self._missing_blocks = missing_blocks

    @property
    def owners(self):
        """Gets the owners of this ArchiveResult.  # noqa: E501


        :return: The owners of this ArchiveResult.  # noqa: E501
        :rtype: list[str]
        """
        return self._owners

    @owners.setter
    def owners(self, owners):
        """Sets the owners of this ArchiveResult.


        :param owners: The owners of this ArchiveResult.  # noqa: E501
        :type: list[str]
        """

        self._owners = owners

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(ArchiveResult, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ArchiveResult):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# __init__.py
from .VMType import VMType
from .Opcode import Opcode
from .EventDecoder import EventDecoder
from .ScriptBuilder import ScriptBuilder
from .Contracts import Contracts
from .Decoder import Decoder
from .EventData import *
from .TypeAuction import TypeAuction
from .VMObject import VMObject



import struct
import base64

from .Decoder import Decoder

def decode_vm_object(str):
    dec = Decoder(str)
    return dec.read_vm_object()

def get_token_event_data(str):
    dec = Decoder(str)
    return {
        'symbol': dec.read_string(),
        'value': dec.read_bigint_accurate(),
        'chainName': dec.read_string()
    }

def get_chain_value_event_data(str):
    dec = Decoder(str)
    return {
        'name': dec.read_string(),
        'value': dec.read_bigint()
    }

def get_transaction_settle_event_data(str):
    dec = Decoder(str)
    return {
        'hash': dec.read_byte(),
        'platform': dec.read_string(),
        'chain': dec.read_string()
    }

def get_gas_event_data(str):
    dec = Decoder(str)
    return {
        'address': dec.read_byte(),
        'price': dec.read_bigint(),
        'amount': dec.read_bigint(),
        'endAmount': 0 if dec.is_end() else dec.read_bigint()
    }

def get_infusion_event_data(str):
    dec = Decoder(str)
    return {
        'baseSymbol': dec.read_string(),
        'TokenID': dec.read_bigint_accurate(),
        'InfusedSymbol': dec.read_string(),
        'InfusedValue': dec.read_bigint_accurate(),
        'ChainName': dec.read_string()
    }

def get_market_event_data(str):
    dec = Decoder(str)
    return {
        'baseSymbol': dec.read_string(),
        'quoteSymbol': dec.read_string(),
        'id': dec.read_bigint_accurate(),
        'amount': dec.read_bigint()
    }

def get_string(str):
    return Decoder(str).read_string()


import requests
import json

class PhantasmaRPC:
	url = None 

	def __init__(self, url):
		self.url = url

	def __JSON_RPC(self, method, params):
		headers = {'content-type': 'application/json'}

		payload = {
				"method": method,
				"params": params,
				"jsonrpc": "2.0",
				"id": 1,
		}
		response = requests.post(self.url, data=json.dumps(payload), headers=headers).json()

		assert response["jsonrpc"]
		assert response["id"] == 1, response
		return response["result"]


	# Account methods
	def getAccount(self, addressText):
		"""
		Returns the account name and balance of given address.
		
		Args:
			addressText: Address of account
		
		"""
		params = [addressText];
		return self.__JSON_RPC("getAccount", params);	
	
	def lookUpName(self, name):
		"""
		Returns the address that owns a given name.
		
		Args:
			name: Name of account
		
		"""
		params = [name];
		return self.__JSON_RPC("lookUpName", params);	

	# Block methods
	def getBlockHeight(self, chainInput):
		"""
		Returns the height of a chain.
		
		Args:
			chainInput: Address or name of chain
		
		"""
		params = [chainInput];
		return self.__JSON_RPC("getBlockHeight", params);	
	
	def getBlockTransactionCountByHash(self, blockHash):
		"""
		Returns the number of transactions of given block hash or error if given hash is invalid or is not found.
		
		Args:
			blockHash: Hash of block
		
		"""
		params = [blockHash];
		return self.__JSON_RPC("getBlockTransactionCountByHash", params);	
	
	def getBlockByHash(self, blockHash):
		"""
		Returns information about a block by hash.
		
		Args:
			blockHash: Hash of block
		
		"""
		params = [blockHash];
		return self.__JSON_RPC("getBlockByHash", params);	
	
	def getRawBlockByHash(self, blockHash):
		"""
		Returns a serialized string, containing information about a block by hash.
		
		Args:
			blockHash: Hash of block
		
		"""
		params = [blockHash];
		return self.__JSON_RPC("getRawBlockByHash", params);	
	
	def getBlockByHeight(self, chainInput, height):
		"""
		Returns information about a block by height and chain.
		
		Args:
			chainInput: Address or name of chain
			height: Height of block
		
		"""
		params = [chainInput, height];
		return self.__JSON_RPC("getBlockByHeight", params);	
	
	def getRawBlockByHeight(self, chainInput, height):
		"""
		Returns a serialized string, in hex format, containing information about a block by height and chain.
		
		Args:
			chainInput: Address or name of chain
			height: Height of block
		
		"""
		params = [chainInput, height];
		return self.__JSON_RPC("getRawBlockByHeight", params);	
	
	def getTransactionByBlockHashAndIndex(self, blockHash, index):
		"""
		Returns the information about a transaction requested by a block hash and transaction index.
		
		Args:
			blockHash: Hash of block
			index: Index of transaction
		
		"""
		params = [blockHash, index];
		return self.__JSON_RPC("getTransactionByBlockHashAndIndex", params);	
	
	def getAddressTransactions(self, addressText, page, pageSize):
		"""
		Returns last X transactions of given address.
		
		Args:
			addressText: Address of account
			page: Index of page to return
			pageSize: Number of items to return per page
		
		"""
		params = [addressText, page, pageSize];
		return self.__JSON_RPC("getAddressTransactions", params);	
	
	def getAddressTransactionCount(self, addressText, chainInput):
		"""
		Get number of transactions in a specific address and chain
		
		Args:
			addressText: Address of account
			chainInput: Name or address of chain, optional
		
		"""
		params = [addressText, chainInput];
		return self.__JSON_RPC("getAddressTransactionCount", params);	
	
	def sendRawTransaction(self, txData):
		"""
		Allows to broadcast a signed operation on the network, but it&apos;s required to build it manually.
		
		Args:
			txData: Serialized transaction bytes, in hexadecimal format
		
		"""
		params = [txData];
		return self.__JSON_RPC("sendRawTransaction", params);	
	
	def invokeRawScript(self, chainInput, scriptData):
		"""
		Allows to invoke script based on network state, without state changes.
		
		Args:
			chainInput: Address or name of chain
			scriptData: Serialized script bytes, in hexadecimal format
		
		"""
		params = [chainInput, scriptData];
		return self.__JSON_RPC("invokeRawScript", params);	
	
	def getTransaction(self, hashText):
		"""
		Returns information about a transaction by hash.
		
		Args:
			hashText: Hash of transaction
		
		"""
		params = [hashText];
		return self.__JSON_RPC("getTransaction", params);	
	
	def cancelTransaction(self, hashText):
		"""
		Removes a pending transaction from the mempool.
		
		Args:
			hashText: Hash of transaction
		
		"""
		params = [hashText];
		return self.__JSON_RPC("cancelTransaction", params);	
	
	def getChains(self):
		"""
		Returns an array of all chains deployed in Phantasma.
		
		Args:
			None
		
		"""
		params = [];
		return self.__JSON_RPC("getChains", params);	
	
	def getTokens(self):
		"""
		Returns an array of tokens deployed in Phantasma.
		
		Args:
			None
		
		"""
		params = [];
		return self.__JSON_RPC("getTokens", params);	
	
	def getToken(self, symbol):
		"""
		Returns info about a specific token deployed in Phantasma.
		
		Args:
			symbol: Token symbol to obtain info
		
		"""
		params = [symbol];
		return self.__JSON_RPC("getToken", params);	
	
	def getTokenData(self, symbol, IDtext):
		"""
		Returns data of a non-fungible token, in hexadecimal format.
		
		Args:
			symbol: Symbol of token
			IDtext: ID of token
		
		"""
		params = [symbol, IDtext];
		return self.__JSON_RPC("getTokenData", params);	
	
	def getApps(self):
		"""
		Returns an array of apps deployed in Phantasma.
		
		Args:
			None
		
		"""
		params = [];
		return self.__JSON_RPC("getApps", params);	
	
	def getTokenTransfers(self, tokenSymbol, page, pageSize):
		"""
		Returns last X transactions of given token.
		
		Args:
			tokenSymbol: Token symbol
			page: Index of page to return
			pageSize: Number of items to return per page
		
		"""
		params = [tokenSymbol, page, pageSize];
		return self.__JSON_RPC("getTokenTransfers", params);	
	
	def getTokenTransferCount(self, tokenSymbol):
		"""
		Returns the number of transaction of a given token.
		
		Args:
			tokenSymbol: Token symbol
		
		"""
		params = [tokenSymbol];
		return self.__JSON_RPC("getTokenTransferCount", params);	
	
	def getTokenBalance(self, addressText, tokenSymbol, chainInput):
		"""
		Returns the balance for a specific token and chain, given an address.
		
		Args:
			addressText: Address of account
			tokenSymbol: Token symbol
			chainInput: Address or name of chain
		
		"""
		params = [addressText, tokenSymbol, chainInput];
		return self.__JSON_RPC("getTokenBalance", params);	
	
	def getAuctionsCount(self, chainAddressOrName, symbol):
		"""
		Returns the number of active auctions.
		
		Args:
			chainAddressOrName: Chain address or name where the market is located
			symbol: Token symbol used as filter
		
		"""
		params = [chainAddressOrName, symbol];
		return self.__JSON_RPC("getAuctionsCount", params);	
	
	def getAuctions(self, chainAddressOrName, symbol, page, pageSize):
		"""
		Returns the auctions available in the market.
		
		Args:
			chainAddressOrName: Chain address or name where the market is located
			symbol: Token symbol used as filter
			page: Index of page to return
			pageSize: Number of items to return per page
		
		"""
		params = [chainAddressOrName, symbol, page, pageSize];
		return self.__JSON_RPC("getAuctions", params);	
	
	def getAuction(self, chainAddressOrName, symbol, IDtext):
		"""
		Returns the auction for a specific token.
		
		Args:
			chainAddressOrName: Chain address or name where the market is located
			symbol: Token symbol
			IDtext: Token ID
		
		"""
		params = [chainAddressOrName, symbol, IDtext];
		return self.__JSON_RPC("getAuction", params);	
	
	def getArchive(self, hashText):
		"""
		Returns info about a specific archive.
		
		Args:
			hashText: Archive hash
		
		"""
		params = [hashText];
		return self.__JSON_RPC("getArchive", params);	
	
	


import struct
import io
from typing import Union

class PBinaryReader:
    def __init__(self, arg1: Union[bytes, bytearray]):
        self.reader = io.BytesIO(arg1)

    @property
    def length(self) -> int:
        current_pos = self.reader.tell()
        self.reader.seek(0, io.SEEK_END)
        length = self.reader.tell()
        self.reader.seek(current_pos, io.SEEK_SET)
        return length

    @property
    def position(self) -> int:
        return self.reader.tell()

    @position.setter
    def position(self, value: int):
        self.reader.seek(value, io.SEEK_SET)

    @property
    def is_end_of_stream(self) -> bool:
        return self.reader.tell() == self.length

    def read_boolean(self) -> bool:
        return struct.unpack('?', self.reader.read(1))[0]

    def read_byte(self) -> int:
        return struct.unpack('B', self.reader.read(1))[0]

    def read_bytes(self, bytes_to_read: int) -> bytes:
        return self.reader.read(bytes_to_read)

    def read_signed_byte(self) -> int:
        return struct.unpack('b', self.reader.read(1))[0]

    def read_short(self) -> int:
        return struct.unpack('<h', self.reader.read(2))[0]

    def read_unsigned_short(self) -> int:
        return struct.unpack('<H', self.reader.read(2))[0]

    def read_int(self) -> int:
        return struct.unpack('<i', self.reader.read(4))[0]

    def read_unsigned_int(self) -> int:
        return struct.unpack('<I', self.reader.read(4))[0]

    def read_long_string(self) -> str:
        length = self.read_var_int()
        return self.read_string_bytes(length)

    def read_long(self) -> int:
        return struct.unpack('<q', self.reader.read(8))[0]

    def read_unsigned_long_string(self) -> str:
        return str(self.read_unsigned_long())

    def read_unsigned_long(self) -> int:
        return struct.unpack('<Q', self.reader.read(8))[0]

    def read_float(self) -> float:
        return struct.unpack('<f', self.reader.read(4))[0]

    def read_double(self) -> float:
        return struct.unpack('<d', self.reader.read(8))[0]

    def read_char(self, encoding: str) -> str:
        return self.reader.read(1).decode(encoding)

    def read_chars(self, characters_to_read: int, encoding: str) -> str:
        return self.reader.read(characters_to_read).decode(encoding)

    def read_big_integer(self) -> int:
        """
        Read a BigInteger. Assumes the BigInteger is unsigned and little-endian encoded.
        """
        length = self.read_var_int()
        if length == 0:
            return 0

        bytes_ = self.read_bytes(length)
        return int.from_bytes(bytes_, byteorder='little', signed=False)

    def read_big_integer_accurate(self) -> str:
        """
        Read a BigInteger accurately as a string.
        """
        length = self.read_var_int()
        if length == 0:
            return "0"

        bytes_ = self.read_bytes(length)
        return str(int.from_bytes(bytes_, byteorder='little', signed=False))

    def read_var_int(self) -> int:
        """
        Read a variable length integer.
        """
        first_byte = self.read_byte()
        if first_byte == 0xfd:
            return struct.unpack('<H', self.read_bytes(2))[0]
        elif first_byte == 0xfe:
            return struct.unpack('<I', self.read_bytes(4))[0]
        elif first_byte == 0xff:
            return struct.unpack('<Q', self.read_bytes(8))[0]
        else:
            return first_byte

    def read_timestamp(self) -> int:
        """
        Read a timestamp. Assumes the timestamp is a 32-bit Unix timestamp.
        """
        bytes_ = self.read_bytes(4)
        return int.from_bytes(bytes_, byteorder='little', signed=False)

    def read_vm_object(self):
        """
        Read a VM object. Implementation depends on the VM's specification.
        """
        # Placeholder for reading a VM object based on your VM's specification.
        pass

    def read_string_bytes(self, num_bytes: int) -> str:
        return self.reader.read(num_bytes).decode()

    # You would need to implement other methods (like readBigInteger, readSignature, etc.) 
    # based on your application logic and the availability of related Python libraries or classes.

# Note: This implementation assumes certain classes and methods are defined elsewhere, like Signature, Timestamp, etc.



from ast import Dict
import base64
from ctypes import Union
from datetime import datetime
from email.headerregistry import Address
from Types.Timestamp import Timestamp
from .VMType import VMType
from typing import Dict, Union

#import big_int  # Assuming big_int is a Python equivalent of big-integer library


class VMObject:
    TimeFormat = "%m/%d/%Y %H:%M:%S"

    def __init__(self):
        self.Type = VMType.NONE
        self.Data = None
        self._local_size = 0

    @property
    def IsEmpty(self):
        return self.Data is None

    def GetChildren(self) -> Union[Dict['VMObject', 'VMObject'], None]:
        return self.Data if self.Type == VMType.Struct else None

    @property
    def Size(self):
        if self.Type == VMType.Object:
            children = self.GetChildren()
            total = sum(entry.Size for entry in children.values())
        else:
            total = self._localSize

        return total

    def AsType(self, type):
        if type == VMType.Bool:
            return self.AsBool()
        elif type == VMType.String:
            return self.AsString()
        elif type == VMType.Bytes:
            return self.AsByteArray()
        elif type == VMType.Number:
            return self.AsNumber()
        elif type == VMType.Timestamp:
            return self.AsTimestamp()
        else:
            raise ValueError("Unsupported VM cast")

    def AsEnum(self, enum_type):
        if not issubclass(enum_type, Enum):
            raise ValueError("enum_type must be an enumerated type")

        if self.Type != VMType.Enum:
            num = self.AsNumber()
            self.Data = int(num)

        return enum_type(self.Data)

    def AsNumber(self):
        if (self.Type in [VMType.Object, VMType.Timestamp]) and isinstance(self.Data, Timestamp):
            return self.Data.Value

        if self.Type == VMType.NoneType:
            return 0

        if self.Type == VMType.String:
            if str(self.Data).isnumeric():
                return int(self.Data)
            else:
                raise Exception(f"Cannot convert String '{self.Data}' to BigInteger.")

        if self.Type == VMType.Bytes:
            bytes_data = bytes(self.Data)
            return int.from_bytes(bytes_data, byteorder='big', signed=False)

        if self.Type == VMType.Enum:
            return int(self.Data)

        if self.Type == VMType.Bool:
            return int(self.Data)

        if self.Type == VMType.Object and isinstance(self.Data, Hash):
            hash_obj = self.Data
            return hash_obj.AsNumber()

        if self.Type != VMType.Number:
            raise Exception(f"Invalid cast: expected number, got {self.Type}")

        return self.Data

    def AsTimestamp(self):
        if self.Type != VMType.Timestamp:
            raise Exception(f"Invalid cast: expected timestamp, got {self.Type}")

        return self.Data
    
    def AsString(self):
        if self.Type == VMType.String:
            return str(self.Data)

        if self.Type == VMType.Number:
            return str(int(self.Data))

        if self.Type == VMType.Bytes:
            return self.Data.decode('utf-8')

        if self.Type == VMType.Enum:
            return str(self.Data)

        if self.Type == VMType.Object:
            if isinstance(self.Data, Address):
                return self.Data.Text
            elif isinstance(self.Data, Hash):
                return str(self.Data)
            else:
                return "Interop:" + type(self.Data).__name__

        if self.Type == VMType.Struct:
            array_type = self.GetArrayType()
            if array_type == VMType.Number:
                children = self.GetChildren()
                result = ""
                for i in range(len(children)):
                    key = VMObject.FromObject(i)
                    val = children[key]
                    ch = chr(int(val.AsNumber()))
                    result += ch
                return result
            else:
                with np.nditer(self.Data, flags=['multi_index'], op_flags=['readonly']) as it:
                    result = bytearray()
                    while not it.finished:
                        index = it.multi_index
                        element = self.Data[index]
                        if isinstance(element, bytes):
                            result.extend(element)
                        else:
                            result.append(element)
                        it.iternext()
                    return base64.b64encode(result).decode('utf-8')

        if self.Type == VMType.Bool:
            return "true" if self.Data else "false"

        if self.Type == VMType.Timestamp:
            date = self.Data
            return date.Value.strftime("%m/%d/%Y %H:%M:%S")

        raise Exception(f"Invalid cast: expected string, got {self.Type}")

    def AsArray(self, type):
        if self.Type != VMType.Struct:
            raise Exception(f"Invalid cast: expected struct, got {self.Type}")

        if self.Data is None:
            return []

        values = self.Data

        if not isinstance(values, dict):
            raise Exception("Invalid struct data")

        result = []

        for i in range(len(values)):
            key = VMObject.FromObject(i)

            if key in values:
                val = values[key]
                result.append(self.CastTo(val, type))

        return result
    
    def GetArrayType(self):
        if self.Type != VMType.Struct:
            return VMType.NoneType

        children = self.GetChildren()

        result = VMType.NoneType

        for i in range(len(children)):
            key = VMObject.FromObject(i)

            if key not in children:
                return VMType.NoneType

            val = children[key]

            if result == VMType.NoneType:
                result = val.Type
            elif val.Type != result:
                return VMType.NoneType

        return result

    def AsByteArray(self):
        if self.Type == VMType.Bytes:
            return bytes(self.Data)

        if self.Type == VMType.Bool:
            return bytes([int(self.Data)])

        if self.Type == VMType.String:
            str_data = self.AsString()
            return str_data.encode('utf-8')

        if self.Type == VMType.Number:
            num_data = self.AsNumber()
            return num_data.to_bytes((num_data.bit_length() + 7) // 8, byteorder='big', signed=True)

        if self.Type == VMType.Enum:
            num_data = int(self.AsNumber())
            return num_data.to_bytes(4, byteorder='big', signed=False)

        if self.Type == VMType.Timestamp:
            time_data = self.AsTimestamp()
            return time_data.Value.to_bytes(8, byteorder='big', signed=False)

        if self.Type == VMType.Struct:
            return self.Serialize()

        if self.Type == VMType.Object:
            serializable = self.Data
            if isinstance(serializable, ISerializable):
                serial_data = serializable.Serialize()[1:]
                return bytearray(serial_data)

            raise Exception("Complex object type can't be converted to bytes")

        raise Exception(f"Invalid cast: expected bytes, got {self.Type}")
    
    def AsAddress(self):
        if self.Type == VMType.String:
            temp = str(self.Data)
            if Address.IsValidAddress(temp):
                return Address.FromText(temp)

        if self.Type == VMType.Bytes:
            temp = bytes(self.Data)

            if len(temp) == Address.LengthInBytes + 1:
                temp = temp[1:]  # Remove the first byte

            if len(temp) == Address.LengthInBytes:
                return Address.FromBytes(temp)

            raise Exception(f"Invalid address size, expected {Address.LengthInBytes} got {len(temp)}")

        if self.Type == VMType.Object and isinstance(self.Data, Address):
            return self.Data

        raise Exception(f"Invalid cast: expected address, got {self.Type}")

    def AsBool(self):
        if self.Type == VMType.Bytes:
            bytes_data = bytes(self.Data)
            if len(bytes_data) == 1:
                return bytes_data[0] != 0

        if self.Type == VMType.Bool:
            return bool(self.Data)

        if self.Type == VMType.Number:
            val = self.AsNumber()
            return val != 0

        raise Exception(f"Invalid cast: expected bool, got {self.Type}")

    def AsStruct(self, struct_type):
        if self.Type == VMType.Object:
            if self.Data is not None and isinstance(self.Data, struct_type):
                return self.Data
            else:
                raise Exception(f"Invalid cast: expected VMObject of type {struct_type.__name__}")

        if self.Type != VMType.Struct:
            raise Exception(f"Invalid cast: expected struct, got {self.Type}")

        if self.Data is None:
            return None

        values = self.Data

        if not isinstance(values, dict):
            raise Exception("Invalid struct data")

        result = struct_type()

        # WARNING: This code is experimental and might not work in every situation.
        for entry in values.items():
            key = entry[0]
            field_name = key.AsString()

            if not field_name:
                raise Exception("Key with null or non-string name found in struct")

            if not hasattr(struct_type, field_name):
                raise Exception("Unknown field: " + field_name)

            field = getattr(struct_type, field_name)
            field_type = field.__annotations__['value']

            field_value = entry[1].ToStruct(field_type) if entry[1].Type == VMType.Struct else entry[1].ToObject()
            field_value = field_value if field_value is not None else ''

            try:
                setattr(result, field_name, field_value)
            except Exception as e:
                raise Exception(f"Could not set field value: {field_name}, {e}")

        return result
    

    @staticmethod
    def ConvertObjectInternal(field_value, field_type):
        if field_type == VMObject:
            # Edge case: happens when calling ToArray<VMObject>()
            return VMObject.FromObject(field_value)
        elif inspect.isclass(field_type) or inspect.isclass(field_value):
            # Check if it's a struct or class and the value is a byte array
            if isinstance(field_value, bytes):
                field_value = Serialization.Unserialize(field_value, field_type)
        elif inspect.isclass(field_type) and issubclass(field_type, enum.Enum):
            # If it's an enum, parse it based on its type
            field_value = field_type(field_value)

        return field_value

    def AsInterop(self, interop_type):
        if interop_type == Hash and self.Type != VMType.Object:
            bytes_data = self.AsByteArray()

            if len(bytes_data) == 33 and bytes_data[0] == 32:
                bytes_data = bytes_data[1:]

            if len(bytes_data) == 32:
                return Hash(bytes_data)

        if self.Type != VMType.Object:
            raise Exception(f"Invalid cast: expected object, got {self.Type}")

        if self.Data is None:
            return None

        if not isinstance(self.Data, interop_type):
            raise Exception("Invalid interop type")

        return self.Data
    
    def SetValue(self, val, vm_type):
        self.Type = vm_type
        self._localSize = len(val) if val is not None else 0

        if vm_type == VMType.Bytes:
            self.Data = val
        elif vm_type == VMType.Number:
            self.Data = BigInteger(0) if val is None or len(val) == 0 else BigInteger(val)
        elif vm_type == VMType.String:
            self.Data = val.decode('utf-8') if val is not None else ""
        elif vm_type == VMType.Enum:
            self.Data = int.from_bytes(val, byteorder='big', signed=False)
        elif vm_type == VMType.Timestamp:
            temp = 0 if val is None else int.from_bytes(val, byteorder='big', signed=False)
            self.Data = Timestamp(temp)
        elif vm_type == VMType.Bool:
            self.Data = bool.from_bytes(val, byteorder='big', signed=False)
        else:
            if isinstance(val, bytes):
                bytes_val = val
                length = len(bytes_val) if bytes_val is not None else 0

                if length == Address.LengthInBytes:
                    self.Data = Address.FromBytes(bytes_val)
                elif length == Hash.Length:
                    self.Data = Hash.FromBytes(bytes_val)
                else:
                    try:
                        self.UnserializeData(bytes_val)
                    except Exception as e:
                        raise Exception("Cannot decode interop object from bytes with length: " + str(length))
            else:
                raise Exception("Cannot set value for VMType: " + vm_type)

        return self
    
    def SetValue(self, val):
        if isinstance(val, BigInteger):
            self.Type = VMType.Number
            self.Data = val
            self._localSize = len(val.to_bytes((val.bit_length() + 7) // 8, byteorder='big', signed=True))
        elif isinstance(val, dict):
            for key in val.keys():
                self.ValidateStructKey(key)
            self.Type = VMType.Struct
            self.Data = val
            self._localSize = 4  # TODO: Update this value
        elif isinstance(val, Hash):
            self.Type = VMType.Object
            self.Data = val
            self._localSize = 4
        elif isinstance(val, Address):
            self.Type = VMType.Object
            self.Data = val
            self._localSize = 4
        elif isinstance(val, Timestamp):
            self.Type = VMType.Timestamp
            self.Data = val
            self._localSize = 4
        elif isinstance(val, str):
            self.Type = VMType.String
            self.Data = val
            self._localSize = len(val)
        elif isinstance(val, bool):
            self.Type = VMType.Bool
            self.Data = val
            self._localSize = 1
        elif isinstance(val, bytes):
            self.Type = VMType.Bytes
            self.Data = val
            self._localSize = len(val)
        elif isinstance(val, enum.Enum):
            self.Type = VMType.Enum
            self.Data = val
            self._localSize = 4
        else:
            val_type = type(val)
            if val is not None and val_type == Timestamp:
                return self.SetValue(val)
            if not inspect.isclass(val_type) or not val_type.__annotations__ or 'value' not in val_type.__annotations__:
                raise Exception(f"Invalid cast: expected struct or class, got {val_type.__name__}")
            self.Type = VMType.Object
            self.Data = val
            self._localSize = 4
        return self

    @staticmethod
    def ValidateStructKey(key):
        if key.Type in [VMType.NoneType, VMType.Struct, VMType.Object]:
            raise Exception(f"Cannot use value of type {key.Type} as key for struct field")

    def SetDefaultValue(self, vm_type):
        self.Type = vm_type
        self._localSize = 1  # TODO: Fixme

        if vm_type == VMType.Bytes:
            self.Data = bytearray()
        elif vm_type == VMType.Number:
            self.Data = BigInteger(0)
        elif vm_type == VMType.String:
            self.Data = ""
        elif vm_type == VMType.Enum:
            self.Data = 0
        elif vm_type == VMType.Timestamp:
            self.Data = Timestamp(0)
        elif vm_type == VMType.Bool:
            self.Data = False
        elif vm_type == VMType.Object:
            self.Data = None
        else:
            raise Exception("Cannot initialize default value for VMType: " + vm_type)

        return self
    
    def SetKey(self, key, obj):
        self.ValidateStructKey(key)
        children = {}

        # NOTE: In Python, we don't need to create a new VMObject instance for the key
        if self.Type == VMType.Struct:
            children = self.GetChildren()
        elif self.Type == VMType.NoneType:
            self.Type = VMType.Struct
            children = {}
            self.Data = children
            self._localSize = 0
        else:
            raise Exception(f"Invalid cast from {self.Type} to struct")

        children[key] = obj

    def GetField(self, key):
        key_obj = VMObject.FromObject(key)
        return self.GetField(key_obj)

    def GetField(self, key):
        if self.Type != VMType.Struct:
            raise Exception(f"Invalid cast: expected struct, got {self.Type}")

        children = self.GetChildren()

        return children.get(key, VMObject())
    
    def GetHashCode(self):
        if self.Type == VMType.Struct:
            hash_value = 2166136261
            children = self.GetChildren()
            for child in children:
                hash_value = (hash_value * 16777619 + child.GetHashCode()) & 0xFFFFFFFF  # Ensure the hash value is within 32-bit bounds
            return hash_value
        else:
            return hash(self.Data) if self.Data is not None else 0

    def Equals(self, other):
        if other is None:
            return False

        if not isinstance(other, VMObject):
            return False

        if self.Type != other.Type:
            return False

        if self.Type == VMType.Struct:
            children = self.GetChildren()
            other_children = other.GetChildren()

            for key, value in children.items():
                if key not in other_children:
                    return False

                if value != other_children[key]:
                    return False

            for key, value in other_children.items():
                if key not in children:
                    return False

            return True
        else:
            return self.Data == other.Data

    def Copy(self, other):
        if other is None or other.Type == VMType.NoneType:
            self.Type = VMType.NoneType
            self.Data = None
            return

        self.Type = other.Type

        if other.Type == VMType.Struct:
            children = {}
            other_children = other.GetChildren()
            for key in other_children.keys():
                temp = VMObject()
                temp.Copy(other_children[key])
                children[key] = temp

            self.Data = children
        else:
            self.Data = other.Data

    def ToString(self):
        if self.Type == VMType.NoneType:
            return "Null"
        elif self.Type == VMType.Struct:
            return "[Struct]"
        elif self.Type == VMType.Bytes:
            return f"[Bytes] => {Base16.Encode(self.Data)}"
        elif self.Type == VMType.Number:
            return f"[Number] => {self.Data}"
        elif self.Type == VMType.Timestamp:
            return f"[Time] => {self.Data.strftime(TimeFormat)}"
        elif self.Type == VMType.String:
            return f"[String] => {self.Data}"
        elif self.Type == VMType.Bool:
            return f"[Bool] => {self.Data}"
        elif self.Type == VMType.Enum:
            return f"[Enum] => {self.Data}"
        elif self.Type == VMType.Object:
            return f"[Object] => {self.Data.__class__.__name__ if self.Data is not None else 'null'}"
        else:
            return "Unknown"

    @staticmethod
    def CastTo(src_obj, type):
        if src_obj.Type == type:
            result = VMObject()
            result.Copy(src_obj)
            return result

        if type == VMType.NoneType:
            return VMObject()

        result = VMObject()

        if type == VMType.String:
            result.SetValue(src_obj.AsString())
        elif type == VMType.Timestamp:
            result.SetValue(src_obj.AsTimestamp())
        elif type == VMType.Bool:
            result.SetValue(src_obj.AsBool())
        elif type == VMType.Bytes:
            result.SetValue(src_obj.AsByteArray())
        elif type == VMType.Number:
            result.SetValue(src_obj.AsNumber())
        elif type == VMType.Struct:
            if src_obj.Type == VMType.String:
                str_value = src_obj.AsString()
                chars = [VMObject.FromObject(BigInteger(ord(c))) for c in str_value]
                return VMObject.FromArray(chars)
            elif src_obj.Type == VMType.Object:
                # You'll need to implement the equivalent of CastViaReflection here.
                # This code depends on your specific class and object structure.
                # Please provide additional information about CastViaReflection to complete this part.
                pass
            else:
                raise Exception(f"Invalid cast: {src_obj.Type} to {type}")
        else:
            raise Exception(f"Invalid cast: {src_obj.Type} to {type}")

        return result
    
    @staticmethod
    def GetVMType(type):
        if type == VMType.Enum:
            return VMType.Enum
        elif type == bool:
            return VMType.Bool
        elif type == str:
            return VMType.String
        elif type == bytes:
            return VMType.Bytes
        elif type == int or type == BigInteger:
            return VMType.Number
        elif type == Timestamp or type == uint:
            return VMType.Timestamp
        elif type.__class__.__name__ == 'Enum':
            return VMType.Enum
        elif type.__class__.__name__ == 'list':
            return VMType.Struct
        elif hasattr(type, '__class__') and (type.__class__.__name__ == 'classobj' or type.__class__.__name__ == 'type'):
            return VMType.Object
        else:
            return VMType.NoneType

    @staticmethod
    def IsVMType(type):
        result = VMObject.GetVMType(type)
        return result != VMType.NoneType

    @staticmethod
    def FromArray(array):
        result = VMObject()
        for i in range(len(array)):
            key = VMObject.FromObject(i)

            temp = array[i]
            val = VMObject.FromObject(temp)

            result.SetKey(key, val)

        return result
    
    @staticmethod
    def FromObject(obj):
        if obj is None:
            # Returns the VMObject equivalent of None
            return VMObject()

        obj_type = type(obj)
        type_enum = VMObject.GetVMType(obj_type)

        if type_enum == VMType.NONE:
            raise Exception("Not a valid object")

        result = VMObject()

        if type_enum == VMType.Bool:
            result.SetValue(bool(obj))
        elif type_enum == VMType.Bytes:
            result.SetValue(bytes(obj), VMType.Bytes)
        elif type_enum == VMType.String:
            result.SetValue(str(obj))
        elif type_enum == VMType.Enum:
            result.SetValue(obj)
        elif type_enum == VMType.Object:
            result.SetValue(obj)
        elif type_enum == VMType.Number:
            if obj_type == int:
                obj = VMObject(int(obj))  # HACK
            result.SetValue(obj)
        elif type_enum == VMType.Timestamp:
            if obj_type == uint:
                obj = Timestamp(uint(obj))  # HACK
            result.SetValue(obj)
        elif type_enum == VMType.Struct:
            if isinstance(obj, list):
                return VMObject.FromArray(obj)
        else:
            return None

        return result

    def ToObject(self):
        if self.Type == VMType.NONE:
            raise Exception("Not a valid object")

        if self.Type == VMType.Bool:
            return self.AsBool()
        elif self.Type == VMType.Bytes:
            return self.AsByteArray()
        elif self.Type == VMType.String:
            return self.AsString()
        elif self.Type == VMType.Number:
            return self.AsNumber()
        elif self.Type == VMType.Timestamp:
            return self.AsTimestamp()
        elif self.Type == VMType.Object:
            return self.Data
        elif self.Type == VMType.Enum:
            return self.Data
        elif self.Type == VMType.Struct:
            if not self.IsEmpty:
                return self.Data
            else:
                raise Exception(f"Cannot cast {self.Type} to object")
        else:
            raise Exception(f"Cannot cast {self.Type} to object")

    def ToObjectWithType(self, type):
        if self.Type == VMType.Struct:
            if isinstance(self.Data, list):
                element_type = type.__args__[0]
                return self.ToArray(element_type)
            else:
                return self.ToStruct(type)
        else:
            temp = self.ToObject()
            return temp
        
    def ToArray(self, array_element_type):
        if self.IsEmpty:
            return []

        if self.Type != VMType.Struct:
            raise Exception("Not a valid source struct")

        children = self.GetChildren()
        max_index = -1
        for child in children:
            if child.Key.Type != VMType.Number:
                raise Exception("Source contains an element with an invalid array index")

            temp = child.Key.AsNumber()

            if temp >= 1024:  # TODO: Use a constant for VM max array size
                raise Exception("Source contains an element with a very large array index")

            index = int(temp)
            if index < 0:
                raise Exception("Source contains an array index with a negative value")

            max_index = max(index, max_index)

        length = max_index + 1
        array = [None] * length

        for child in children:
            temp = child.Key.AsNumber()
            index = int(temp)

            val = child.Value.ToObjectWithType(array_element_type)

            val = ConvertObjectInternal(val, array_element_type)

            array[index] = val

        return array
    
    def ToStruct(self, struct_type):
        if self.Type != VMType.Struct:
            raise Exception("Not a valid source struct")

        if not inspect.isclass(struct_type) or not hasattr(struct_type, '__dict__'):
            raise Exception("Not a valid destination struct")

        struct_fields = struct_type.__dict__
        result = struct_type()

        for field_name, field_type in struct_fields.items():
            key = VMObject.FromObject(field_name)

            if key in self.Data:
                val = self.Data[key].ToObjectWithType(field_type)
            else:
                if inspect.isclass(field_type) or inspect.isclass(field_type.__class__):
                    val = None
                else:
                    raise Exception("Field not present in source struct: " + field_name)

            # Here we check if the types mismatch.
            # In case of getting a bytes object instead of an object, we try deserializing the bytes in a different approach.
            # This should not be necessary often.
            if val is not None and field_type != bytes and isinstance(val, bytes):
                if issubclass(field_type, ISerializable):
                    temp = field_type()
                    bytes = val
                    stream = io.BytesIO(bytes)
                    reader = BinaryReader(stream)
                    temp.UnserializeData(reader)
                    val = temp

            # HACK: Allows treating uints as enums, without this it is impossible to transform between C# objects and VM objects.
            if hasattr(field_type, '__enum__') and not isinstance(val, Enum):
                val = field_type(val)

            setattr(result, field_name, val)

        return result
    
    def CastViaReflection(src_obj, level=0, dont_convert_serializables=True):
        src_type = type(src_obj)

        if isinstance(src_obj, pyarray.Array):
            children = {}

            for i, val in enumerate(src_obj):
                key = VMObject()
                key.SetValue(i)
                vm_val = CastViaReflection(val, level + 1)
                children[key] = vm_val

            result = VMObject()
            result.SetValue(children)
            return result
        else:
            target_type = GetVMType(src_type)

            result = None
            is_known_type = src_type in [int, bool, str, bytes, BigInteger, Timestamp]

            if not is_known_type and dont_convert_serializables and issubclass(src_type, ISerializable):
                is_known_type = True

            if is_known_type is False and src_type.__class__.__name__ not in ["type", "EnumMeta"]:
                children = {}

                fields = [attr for attr in dir(src_obj) if not callable(getattr(src_obj, attr)) and not attr.startswith("__")]

                if len(fields) > 0:
                    for field_name in fields:
                        key = VMObject()
                        key.SetValue(field_name)
                        ValidateStructKey(key)

                        val = getattr(src_obj, field_name)
                        vm_val = CastViaReflection(val, level + 1, True)
                        children[key] = vm_val

                    result = VMObject()
                    result.SetValue(children)
                    return result

            if not is_known_type:
                result = VMObject.FromObject(src_obj)
                if result is not None:
                    return result

            raise Exception(f"Invalid cast: Interop.{src_type.__name__} to vm object")
    
    def SerializeData(writer):
        writer.write(bytes([self.Type]))
        
        if self.Type == VMType.NONE:
            return

        data_type = type(self.Data)

        if self.Type == VMType.Struct:
            children = self.GetChildren()
            writer.write_var_int(len(children))
            
            for entry in children:
                entry[0].SerializeData(writer)
                entry[1].SerializeData(writer)
        
        elif self.Type == VMType.Object:
            obj = self.Data
            if isinstance(obj, ISerializable):
                bytes_data = Serialization.Serialize(obj)
                writer.write_byte_array(bytes_data)
            elif isinstance(obj, list):
                writer.write_var_int(len(obj))
                for item in obj:
                    obj2 = CastViaReflection(item, 0, False)
                    obj2.SerializeData(writer)
            else:
                raise Exception(f"Objects of type {data_type.__name__} cannot be serialized")
        
        elif self.Type == VMType.Enum:
            temp2 = int(self.Data) if isinstance(self.Data, Enum) else self.Data
            writer.write_var_int(temp2)
        
        else:
            if self.Type == VMType.String:
                data = self.Data.encode('utf-8')
                writer.write_var_int(len(data))
                writer.write(data)
            elif self.Type == VMType.Number:
                temp1 = str(self.Data)
                temp2 = int(temp1) if isinstance(self.Data, int) else self.Data
                writer.write_var_int(temp2)
            else:
                Serialization.Serialize(writer, self.Data)

    @staticmethod
    def FromBytes(bytes):
        result = VMObject()
        result.UnserializeData(bytes)
        return result

    @staticmethod
    def FromStruct(obj):
        return CastViaReflection(obj, 0, False)
    
    def UnserializeData(bytes_data):
        stream = io.BytesIO(bytes_data)
        reader = BinaryReader(stream)
        self.Type = VMType(reader.read_byte())
        
        if self.Type == VMType.Bool:
            self.Data = Serialization.Unserialize[bool](reader)
            self._localSize = 1
        elif self.Type == VMType.Bytes:
            self.Data = Serialization.Unserialize[bytearray](reader)
            self._localSize = len(self.Data)
        elif self.Type == VMType.Number:
            self.Data = Serialization.Unserialize[bigint](reader)
            self._localSize = 32
        elif self.Type == VMType.Timestamp:
            self.Data = Serialization.Unserialize[Timestamp](reader)
            self._localSize = 8
        elif self.Type == VMType.String:
            self.Data = Serialization.Unserialize[str](reader)
            self._localSize = len(self.Data) if self.Data is not None else 0
        elif self.Type == VMType.Struct:
            child_count = reader.read_var_int()
            self._localSize = child_count if child_count is not None else 0
            children = {}
            
            while child_count > 0:
                key = VMObject()
                key.UnserializeData(reader)
                ValidateStructKey(key)
                
                val = VMObject()
                val.UnserializeData(reader)
                children[key] = val
                child_count -= 1
            
            self.Data = children
        elif self.Type == VMType.Object:
            bytes_data = reader.read_byte_array()
            self._localSize = len(bytes_data) if bytes_data is not None else 0
            
            if len(bytes_data) == 35:
                addr = Serialization.Unserialize[Address](bytes_data)
                self.Data = addr
                self.Type = VMType.Object
            else:
                self.Type = VMType.Bytes
                self.Data = bytes_data
        elif self.Type == VMType.Enum:
            self.Type = VMType.Enum
            self.Data = reader.read_var_int()
            self._localSize = 1
        elif self.Type == VMType.NONE:
            self.Type = VMType.NONE
            self.Data = None
        else:
            raise Exception(f"Invalid unserialize: type {self.Type}")
    
    def __str__(self):
        if self.Type == VMType.String:
            return str(self.Data)
        elif self.Type == VMType.Number:
            return str(self.Data)
        elif self.Type == VMType.Bytes:
            return base64.b64encode(self.Data).decode('utf-8')  # or any other suitable representation
        elif self.Type == VMType.Enum:
            return str(self.Data)
        elif self.Type == VMType.Bool:
            return "true" if self.Data else "false"
        elif self.Type == VMType.Timestamp:
            return str(self.Data)  # Assuming Data holds a datetime object
        elif self.Type == VMType.Struct or self.Type == VMType.Object:
            # Represent Struct or Object as a string (customize as needed)
            return f"Object of type {self.Type} with data: {self.Data}"
        else:
            raise ValueError(f"Invalid cast: expected string, got {self.Type}")



from abc import ABC, abstractmethod
from typing import List
from Interfaces.SignatureKind import SignatureKind
from Types import Address
from Types.Extensions import PBinaryReader, PBinaryWriter

class Signature(ABC):
    @property
    @abstractmethod
    def Bytes(self) -> bytes:
        pass

    @property
    @abstractmethod
    def Kind(self) -> SignatureKind:
        pass

    @abstractmethod
    def SerializeData(self, writer: PBinaryWriter):
        pass

    @abstractmethod
    def UnserializeData(self, reader: PBinaryReader):
        pass

    @abstractmethod
    def VerifyMultiple(self, message: bytes, addresses: List[Address]) -> bool:
        pass

    def Verify(self, message: bytes, address: Address) -> bool:
        return self.VerifyMultiple(message, [address])

    def ToByteArray(self) -> bytes:
        stream = bytearray(64)
        writer = PBinaryWriter(stream)
        self.SerializeData(writer)
        return bytes(stream)


from typing import Callable, Optional
import numpy as np
from .Signature import Signature

class IKeyPair:
    def __init__(self, private_key: bytes, public_key: bytes):
        self.private_key = private_key
        self.public_key = public_key

    def sign(self, msg: bytes, custom_sign_function: Optional[Callable[[bytes, bytes, bytes], bytes]] = None) -> Signature:
        if custom_sign_function is not None:
            signature_bytes = custom_sign_function(msg, self.private_key, self.public_key)
        else:
            signature_bytes = self.default_sign_function(msg, self.private_key, self.public_key)
        return self.create_signature(signature_bytes)

    def default_sign_function(self, message: bytes, private_key: bytes, public_key: bytes) -> bytes:
        # Implement the default signing logic here
        pass

    def create_signature(self, signature_bytes: bytes) -> Signature:
        # Create and return a Signature object from the signature bytes
        pass



# coding: utf-8

"""
    Phantasma API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class AuctionResult(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'creator_address': 'str',
        'chain_address': 'str',
        'start_date': 'int',
        'end_date': 'int',
        'base_symbol': 'str',
        'quote_symbol': 'str',
        'token_id': 'str',
        'price': 'str',
        'end_price': 'str',
        'extension_period': 'str',
        'type': 'str',
        'rom': 'str',
        'ram': 'str',
        'listing_fee': 'str',
        'current_winner': 'str'
    }

    attribute_map = {
        'creator_address': 'creatorAddress',
        'chain_address': 'chainAddress',
        'start_date': 'startDate',
        'end_date': 'endDate',
        'base_symbol': 'baseSymbol',
        'quote_symbol': 'quoteSymbol',
        'token_id': 'tokenId',
        'price': 'price',
        'end_price': 'endPrice',
        'extension_period': 'extensionPeriod',
        'type': 'type',
        'rom': 'rom',
        'ram': 'ram',
        'listing_fee': 'listingFee',
        'current_winner': 'currentWinner'
    }

    def __init__(self, creator_address=None, chain_address=None, start_date=None, end_date=None, base_symbol=None, quote_symbol=None, token_id=None, price=None, end_price=None, extension_period=None, type=None, rom=None, ram=None, listing_fee=None, current_winner=None):  # noqa: E501
        """AuctionResult - a model defined in Swagger"""  # noqa: E501
        self._creator_address = None
        self._chain_address = None
        self._start_date = None
        self._end_date = None
        self._base_symbol = None
        self._quote_symbol = None
        self._token_id = None
        self._price = None
        self._end_price = None
        self._extension_period = None
        self._type = None
        self._rom = None
        self._ram = None
        self._listing_fee = None
        self._current_winner = None
        self.discriminator = None
        if creator_address is not None:
            self.creator_address = creator_address
        if chain_address is not None:
            self.chain_address = chain_address
        if start_date is not None:
            self.start_date = start_date
        if end_date is not None:
            self.end_date = end_date
        if base_symbol is not None:
            self.base_symbol = base_symbol
        if quote_symbol is not None:
            self.quote_symbol = quote_symbol
        if token_id is not None:
            self.token_id = token_id
        if price is not None:
            self.price = price
        if end_price is not None:
            self.end_price = end_price
        if extension_period is not None:
            self.extension_period = extension_period
        if type is not None:
            self.type = type
        if rom is not None:
            self.rom = rom
        if ram is not None:
            self.ram = ram
        if listing_fee is not None:
            self.listing_fee = listing_fee
        if current_winner is not None:
            self.current_winner = current_winner

    @property
    def creator_address(self):
        """Gets the creator_address of this AuctionResult.  # noqa: E501


        :return: The creator_address of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._creator_address

    @creator_address.setter
    def creator_address(self, creator_address):
        """Sets the creator_address of this AuctionResult.


        :param creator_address: The creator_address of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._creator_address = creator_address

    @property
    def chain_address(self):
        """Gets the chain_address of this AuctionResult.  # noqa: E501


        :return: The chain_address of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._chain_address

    @chain_address.setter
    def chain_address(self, chain_address):
        """Sets the chain_address of this AuctionResult.


        :param chain_address: The chain_address of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._chain_address = chain_address

    @property
    def start_date(self):
        """Gets the start_date of this AuctionResult.  # noqa: E501


        :return: The start_date of this AuctionResult.  # noqa: E501
        :rtype: int
        """
        return self._start_date

    @start_date.setter
    def start_date(self, start_date):
        """Sets the start_date of this AuctionResult.


        :param start_date: The start_date of this AuctionResult.  # noqa: E501
        :type: int
        """

        self._start_date = start_date

    @property
    def end_date(self):
        """Gets the end_date of this AuctionResult.  # noqa: E501


        :return: The end_date of this AuctionResult.  # noqa: E501
        :rtype: int
        """
        return self._end_date

    @end_date.setter
    def end_date(self, end_date):
        """Sets the end_date of this AuctionResult.


        :param end_date: The end_date of this AuctionResult.  # noqa: E501
        :type: int
        """

        self._end_date = end_date

    @property
    def base_symbol(self):
        """Gets the base_symbol of this AuctionResult.  # noqa: E501


        :return: The base_symbol of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._base_symbol

    @base_symbol.setter
    def base_symbol(self, base_symbol):
        """Sets the base_symbol of this AuctionResult.


        :param base_symbol: The base_symbol of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._base_symbol = base_symbol

    @property
    def quote_symbol(self):
        """Gets the quote_symbol of this AuctionResult.  # noqa: E501


        :return: The quote_symbol of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._quote_symbol

    @quote_symbol.setter
    def quote_symbol(self, quote_symbol):
        """Sets the quote_symbol of this AuctionResult.


        :param quote_symbol: The quote_symbol of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._quote_symbol = quote_symbol

    @property
    def token_id(self):
        """Gets the token_id of this AuctionResult.  # noqa: E501


        :return: The token_id of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._token_id

    @token_id.setter
    def token_id(self, token_id):
        """Sets the token_id of this AuctionResult.


        :param token_id: The token_id of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._token_id = token_id

    @property
    def price(self):
        """Gets the price of this AuctionResult.  # noqa: E501


        :return: The price of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._price

    @price.setter
    def price(self, price):
        """Sets the price of this AuctionResult.


        :param price: The price of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._price = price

    @property
    def end_price(self):
        """Gets the end_price of this AuctionResult.  # noqa: E501


        :return: The end_price of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._end_price

    @end_price.setter
    def end_price(self, end_price):
        """Sets the end_price of this AuctionResult.


        :param end_price: The end_price of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._end_price = end_price

    @property
    def extension_period(self):
        """Gets the extension_period of this AuctionResult.  # noqa: E501


        :return: The extension_period of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._extension_period

    @extension_period.setter
    def extension_period(self, extension_period):
        """Sets the extension_period of this AuctionResult.


        :param extension_period: The extension_period of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._extension_period = extension_period

    @property
    def type(self):
        """Gets the type of this AuctionResult.  # noqa: E501


        :return: The type of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this AuctionResult.


        :param type: The type of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._type = type

    @property
    def rom(self):
        """Gets the rom of this AuctionResult.  # noqa: E501


        :return: The rom of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._rom

    @rom.setter
    def rom(self, rom):
        """Sets the rom of this AuctionResult.


        :param rom: The rom of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._rom = rom

    @property
    def ram(self):
        """Gets the ram of this AuctionResult.  # noqa: E501


        :return: The ram of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._ram

    @ram.setter
    def ram(self, ram):
        """Sets the ram of this AuctionResult.


        :param ram: The ram of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._ram = ram

    @property
    def listing_fee(self):
        """Gets the listing_fee of this AuctionResult.  # noqa: E501


        :return: The listing_fee of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._listing_fee

    @listing_fee.setter
    def listing_fee(self, listing_fee):
        """Sets the listing_fee of this AuctionResult.


        :param listing_fee: The listing_fee of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._listing_fee = listing_fee

    @property
    def current_winner(self):
        """Gets the current_winner of this AuctionResult.  # noqa: E501


        :return: The current_winner of this AuctionResult.  # noqa: E501
        :rtype: str
        """
        return self._current_winner

    @current_winner.setter
    def current_winner(self, current_winner):
        """Sets the current_winner of this AuctionResult.


        :param current_winner: The current_winner of this AuctionResult.  # noqa: E501
        :type: str
        """

        self._current_winner = current_winner

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(AuctionResult, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, AuctionResult):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    Phantasma API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class NetInfoPeerConnectionStatusMonitor(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'start': 'str',
        'bytes': 'str',
        'samples': 'str',
        'inst_rate': 'str',
        'cur_rate': 'str',
        'avg_rate': 'str',
        'peak_rate': 'str',
        'bytes_rem': 'str',
        'duration': 'str',
        'idle': 'str',
        'time_rem': 'str',
        'progress': 'int',
        'active': 'bool'
    }

    attribute_map = {
        'start': 'start',
        'bytes': 'bytes',
        'samples': 'samples',
        'inst_rate': 'instRate',
        'cur_rate': 'curRate',
        'avg_rate': 'avgRate',
        'peak_rate': 'peakRate',
        'bytes_rem': 'bytesRem',
        'duration': 'duration',
        'idle': 'idle',
        'time_rem': 'timeRem',
        'progress': 'progress',
        'active': 'active'
    }

    def __init__(self, start=None, bytes=None, samples=None, inst_rate=None, cur_rate=None, avg_rate=None, peak_rate=None, bytes_rem=None, duration=None, idle=None, time_rem=None, progress=None, active=None):  # noqa: E501
        """NetInfoPeerConnectionStatusMonitor - a model defined in Swagger"""  # noqa: E501
        self._start = None
        self._bytes = None
        self._samples = None
        self._inst_rate = None
        self._cur_rate = None
        self._avg_rate = None
        self._peak_rate = None
        self._bytes_rem = None
        self._duration = None
        self._idle = None
        self._time_rem = None
        self._progress = None
        self._active = None
        self.discriminator = None
        if start is not None:
            self.start = start
        if bytes is not None:
            self.bytes = bytes
        if samples is not None:
            self.samples = samples
        if inst_rate is not None:
            self.inst_rate = inst_rate
        if cur_rate is not None:
            self.cur_rate = cur_rate
        if avg_rate is not None:
            self.avg_rate = avg_rate
        if peak_rate is not None:
            self.peak_rate = peak_rate
        if bytes_rem is not None:
            self.bytes_rem = bytes_rem
        if duration is not None:
            self.duration = duration
        if idle is not None:
            self.idle = idle
        if time_rem is not None:
            self.time_rem = time_rem
        if progress is not None:
            self.progress = progress
        if active is not None:
            self.active = active

    @property
    def start(self):
        """Gets the start of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The start of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: str
        """
        return self._start

    @start.setter
    def start(self, start):
        """Sets the start of this NetInfoPeerConnectionStatusMonitor.


        :param start: The start of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: str
        """

        self._start = start

    @property
    def bytes(self):
        """Gets the bytes of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The bytes of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: str
        """
        return self._bytes

    @bytes.setter
    def bytes(self, bytes):
        """Sets the bytes of this NetInfoPeerConnectionStatusMonitor.


        :param bytes: The bytes of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: str
        """

        self._bytes = bytes

    @property
    def samples(self):
        """Gets the samples of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The samples of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: str
        """
        return self._samples

    @samples.setter
    def samples(self, samples):
        """Sets the samples of this NetInfoPeerConnectionStatusMonitor.


        :param samples: The samples of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: str
        """

        self._samples = samples

    @property
    def inst_rate(self):
        """Gets the inst_rate of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The inst_rate of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: str
        """
        return self._inst_rate

    @inst_rate.setter
    def inst_rate(self, inst_rate):
        """Sets the inst_rate of this NetInfoPeerConnectionStatusMonitor.


        :param inst_rate: The inst_rate of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: str
        """

        self._inst_rate = inst_rate

    @property
    def cur_rate(self):
        """Gets the cur_rate of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The cur_rate of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: str
        """
        return self._cur_rate

    @cur_rate.setter
    def cur_rate(self, cur_rate):
        """Sets the cur_rate of this NetInfoPeerConnectionStatusMonitor.


        :param cur_rate: The cur_rate of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: str
        """

        self._cur_rate = cur_rate

    @property
    def avg_rate(self):
        """Gets the avg_rate of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The avg_rate of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: str
        """
        return self._avg_rate

    @avg_rate.setter
    def avg_rate(self, avg_rate):
        """Sets the avg_rate of this NetInfoPeerConnectionStatusMonitor.


        :param avg_rate: The avg_rate of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: str
        """

        self._avg_rate = avg_rate

    @property
    def peak_rate(self):
        """Gets the peak_rate of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The peak_rate of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: str
        """
        return self._peak_rate

    @peak_rate.setter
    def peak_rate(self, peak_rate):
        """Sets the peak_rate of this NetInfoPeerConnectionStatusMonitor.


        :param peak_rate: The peak_rate of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: str
        """

        self._peak_rate = peak_rate

    @property
    def bytes_rem(self):
        """Gets the bytes_rem of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The bytes_rem of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: str
        """
        return self._bytes_rem

    @bytes_rem.setter
    def bytes_rem(self, bytes_rem):
        """Sets the bytes_rem of this NetInfoPeerConnectionStatusMonitor.


        :param bytes_rem: The bytes_rem of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: str
        """

        self._bytes_rem = bytes_rem

    @property
    def duration(self):
        """Gets the duration of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The duration of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: str
        """
        return self._duration

    @duration.setter
    def duration(self, duration):
        """Sets the duration of this NetInfoPeerConnectionStatusMonitor.


        :param duration: The duration of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: str
        """

        self._duration = duration

    @property
    def idle(self):
        """Gets the idle of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The idle of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: str
        """
        return self._idle

    @idle.setter
    def idle(self, idle):
        """Sets the idle of this NetInfoPeerConnectionStatusMonitor.


        :param idle: The idle of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: str
        """

        self._idle = idle

    @property
    def time_rem(self):
        """Gets the time_rem of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The time_rem of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: str
        """
        return self._time_rem

    @time_rem.setter
    def time_rem(self, time_rem):
        """Sets the time_rem of this NetInfoPeerConnectionStatusMonitor.


        :param time_rem: The time_rem of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: str
        """

        self._time_rem = time_rem

    @property
    def progress(self):
        """Gets the progress of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The progress of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: int
        """
        return self._progress

    @progress.setter
    def progress(self, progress):
        """Sets the progress of this NetInfoPeerConnectionStatusMonitor.


        :param progress: The progress of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: int
        """

        self._progress = progress

    @property
    def active(self):
        """Gets the active of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501


        :return: The active of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :rtype: bool
        """
        return self._active

    @active.setter
    def active(self, active):
        """Sets the active of this NetInfoPeerConnectionStatusMonitor.


        :param active: The active of this NetInfoPeerConnectionStatusMonitor.  # noqa: E501
        :type: bool
        """

        self._active = active

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(NetInfoPeerConnectionStatusMonitor, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, NetInfoPeerConnectionStatusMonitor):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    Phantasma API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class PlatformResult(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'platform': 'str',
        'chain': 'str',
        'fuel': 'str',
        'tokens': 'list[str]',
        'interop': 'list[InteropResult]'
    }

    attribute_map = {
        'platform': 'platform',
        'chain': 'chain',
        'fuel': 'fuel',
        'tokens': 'tokens',
        'interop': 'interop'
    }

    def __init__(self, platform=None, chain=None, fuel=None, tokens=None, interop=None):  # noqa: E501
        """PlatformResult - a model defined in Swagger"""  # noqa: E501
        self._platform = None
        self._chain = None
        self._fuel = None
        self._tokens = None
        self._interop = None
        self.discriminator = None
        if platform is not None:
            self.platform = platform
        if chain is not None:
            self.chain = chain
        if fuel is not None:
            self.fuel = fuel
        if tokens is not None:
            self.tokens = tokens
        if interop is not None:
            self.interop = interop

    @property
    def platform(self):
        """Gets the platform of this PlatformResult.  # noqa: E501


        :return: The platform of this PlatformResult.  # noqa: E501
        :rtype: str
        """
        return self._platform

    @platform.setter
    def platform(self, platform):
        """Sets the platform of this PlatformResult.


        :param platform: The platform of this PlatformResult.  # noqa: E501
        :type: str
        """

        self._platform = platform

    @property
    def chain(self):
        """Gets the chain of this PlatformResult.  # noqa: E501


        :return: The chain of this PlatformResult.  # noqa: E501
        :rtype: str
        """
        return self._chain

    @chain.setter
    def chain(self, chain):
        """Sets the chain of this PlatformResult.


        :param chain: The chain of this PlatformResult.  # noqa: E501
        :type: str
        """

        self._chain = chain

    @property
    def fuel(self):
        """Gets the fuel of this PlatformResult.  # noqa: E501


        :return: The fuel of this PlatformResult.  # noqa: E501
        :rtype: str
        """
        return self._fuel

    @fuel.setter
    def fuel(self, fuel):
        """Sets the fuel of this PlatformResult.


        :param fuel: The fuel of this PlatformResult.  # noqa: E501
        :type: str
        """

        self._fuel = fuel

    @property
    def tokens(self):
        """Gets the tokens of this PlatformResult.  # noqa: E501


        :return: The tokens of this PlatformResult.  # noqa: E501
        :rtype: list[str]
        """
        return self._tokens

    @tokens.setter
    def tokens(self, tokens):
        """Sets the tokens of this PlatformResult.


        :param tokens: The tokens of this PlatformResult.  # noqa: E501
        :type: list[str]
        """

        self._tokens = tokens

    @property
    def interop(self):
        """Gets the interop of this PlatformResult.  # noqa: E501


        :return: The interop of this PlatformResult.  # noqa: E501
        :rtype: list[InteropResult]
        """
        return self._interop

    @interop.setter
    def interop(self, interop):
        """Sets the interop of this PlatformResult.


        :param interop: The interop of this PlatformResult.  # noqa: E501
        :type: list[InteropResult]
        """

        self._interop = interop

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(PlatformResult, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, PlatformResult):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    Phantasma API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class NetInfoPeer(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'node_info': 'NodeInfo',
        'is_outbound': 'bool',
        'connection_status': 'NetInfoPeerConnectionStatus',
        'remote_ip': 'str'
    }

    attribute_map = {
        'node_info': 'nodeInfo',
        'is_outbound': 'isOutbound',
        'connection_status': 'connectionStatus',
        'remote_ip': 'remoteIp'
    }

    def __init__(self, node_info=None, is_outbound=None, connection_status=None, remote_ip=None):  # noqa: E501
        """NetInfoPeer - a model defined in Swagger"""  # noqa: E501
        self._node_info = None
        self._is_outbound = None
        self._connection_status = None
        self._remote_ip = None
        self.discriminator = None
        if node_info is not None:
            self.node_info = node_info
        if is_outbound is not None:
            self.is_outbound = is_outbound
        if connection_status is not None:
            self.connection_status = connection_status
        if remote_ip is not None:
            self.remote_ip = remote_ip

    @property
    def node_info(self):
        """Gets the node_info of this NetInfoPeer.  # noqa: E501


        :return: The node_info of this NetInfoPeer.  # noqa: E501
        :rtype: NodeInfo
        """
        return self._node_info

    @node_info.setter
    def node_info(self, node_info):
        """Sets the node_info of this NetInfoPeer.


        :param node_info: The node_info of this NetInfoPeer.  # noqa: E501
        :type: NodeInfo
        """

        self._node_info = node_info

    @property
    def is_outbound(self):
        """Gets the is_outbound of this NetInfoPeer.  # noqa: E501


        :return: The is_outbound of this NetInfoPeer.  # noqa: E501
        :rtype: bool
        """
        return self._is_outbound

    @is_outbound.setter
    def is_outbound(self, is_outbound):
        """Sets the is_outbound of this NetInfoPeer.


        :param is_outbound: The is_outbound of this NetInfoPeer.  # noqa: E501
        :type: bool
        """

        self._is_outbound = is_outbound

    @property
    def connection_status(self):
        """Gets the connection_status of this NetInfoPeer.  # noqa: E501


        :return: The connection_status of this NetInfoPeer.  # noqa: E501
        :rtype: NetInfoPeerConnectionStatus
        """
        return self._connection_status

    @connection_status.setter
    def connection_status(self, connection_status):
        """Sets the connection_status of this NetInfoPeer.


        :param connection_status: The connection_status of this NetInfoPeer.  # noqa: E501
        :type: NetInfoPeerConnectionStatus
        """

        self._connection_status = connection_status

    @property
    def remote_ip(self):
        """Gets the remote_ip of this NetInfoPeer.  # noqa: E501


        :return: The remote_ip of this NetInfoPeer.  # noqa: E501
        :rtype: str
        """
        return self._remote_ip

    @remote_ip.setter
    def remote_ip(self, remote_ip):
        """Sets the remote_ip of this NetInfoPeer.


        :param remote_ip: The remote_ip of this NetInfoPeer.  # noqa: E501
        :type: str
        """

        self._remote_ip = remote_ip

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(NetInfoPeer, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, NetInfoPeer):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



class IStack:
    def __init__(self):
        self._stack = []

    def push(self, item):
        """
        Push an item onto the stack.
        """
        self._stack.append(item)

    def pop(self):
        """
        Pop an item off the stack.
        """
        return self._stack.pop() if self._stack else None

    def peek(self):
        """
        Peek at the top item of the stack without removing it.
        """
        return self._stack[-1] if self._stack else None

    def size(self):
        """
        Get the number of items in the stack.
        """
        return len(self._stack)


from enum import Enum


class TokenTrigger(Enum):
    OnMint = 0
    OnBurn = 1
    OnSend = 2
    OnReceive = 3
    OnInfuse = 4
    OnUpgrade = 5
    OnSeries = 6
    OnWrite = 7
    OnMigrate = 8
    OnKill = 9


import struct
from datetime import datetime
from typing import Union, List

from Interfaces import Signature, SignatureKind

# Assuming Signature and SignatureKind are already defined in your Python project

class PBinaryWriter:
    def __init__(self, arg1: Union[None, bytearray, bytes] = None):
        if arg1 is None:
            self._buffer = bytearray()
        elif isinstance(arg1, (bytearray, bytes)):
            self._buffer = bytearray(arg1)
        else:
            raise TypeError("Invalid argument type")

    @property
    def length(self) -> int:
        return len(self._buffer)

    @property
    def position(self) -> int:
        return len(self._buffer)

    def write_boolean(self, value: bool) -> None:
        self._buffer += struct.pack('?', value)

    def write_byte(self, value: int) -> None:
        self._buffer += struct.pack('B', value)

    def write_signed_byte(self, value: int) -> None:
        self._buffer += struct.pack('b', value)

    def write_short(self, value: int) -> None:
        self._buffer += struct.pack('<h', value)

    def write_unsigned_short(self, value: int) -> None:
        self._buffer += struct.pack('<H', value)

    def write_int(self, value: int) -> None:
        self._buffer += struct.pack('<i', value)

    def write_unsigned_int(self, value: int) -> None:
        self._buffer += struct.pack('<I', value)

    def write_long(self, value: Union[str, int]) -> None:
        self._buffer += struct.pack('<q', int(value))

    def write_unsigned_long(self, value: Union[str, int]) -> None:
        self._buffer += struct.pack('<Q', int(value))

    def write_float(self, value: float) -> None:
        self._buffer += struct.pack('<f', value)

    def write_double(self, value: float) -> None:
        self._buffer += struct.pack('<d', value)

    # ... more methods ...

    def to_byte_array(self) -> bytearray:
        return self._buffer

    def write_var_int(self, value: int) -> None:
        if value < 0:
            raise ValueError("negative value invalid")
        if value < 0xfd:
            self.write_byte(value)
        elif value <= 0xffff:
            self.write_byte(0xfd)
            self.write_unsigned_short(value)
        elif value <= 0xffffffff:
            self.write_byte(0xfe)
            self.write_unsigned_int(value)
        else:
            self.write_byte(0xff)
            self.write_unsigned_long(value)

    def write_chars(self, characters: Union[str, List[int]], encoding: str) -> None:
        if isinstance(characters, str):
            self._buffer += characters.encode(encoding)
        elif isinstance(characters, list):
            for char in characters:
                self._buffer += struct.pack('c', char.to_bytes(1, 'little'))

    def clear(self) -> None:
        self._buffer.clear()

    def to_uint8_array(self) -> bytes:
        return bytes(self._buffer)

    def append_bytes(self, bytes_: List[int]) -> None:
        for byte in bytes_:
            if byte is not None and byte != float('nan'):
                self.write_byte(byte)

    def write_enum(self, value: int) -> None:
        bytes_ = [(value >> (8 * i)) & 0xff for i in range(4)]
        self.append_bytes(bytes_)

    def write_bytes(self, bytes_: List[int]) -> None:
        for byte in bytes_:
            self.write_byte(byte)

    def write_timestamp(self, timestamp: datetime) -> None:
        unix_time = int(timestamp.timestamp())
        self.write_unsigned_int(unix_time)

    def write_date_time(self, date_time: datetime) -> None:
        unix_time = int(date_time.timestamp())
        self.write_unsigned_int(unix_time)

    def raw_string(self, value: str) -> List[int]:
        return [ord(char) for char in value]

    def write_byte_array(self, bytes_: Union[List[int], bytes, bytearray]) -> None:
        if isinstance(bytes_, (bytes, bytearray)):
            bytes_ = list(bytes_)
        self.write_var_int(len(bytes_))
        self.write_bytes(bytes_)

    def write_string(self, text: str) -> None:
        bytes_ = self.raw_string(text)
        self.write_var_int(len(bytes_))
        self.write_bytes(bytes_)

    def emit_uint32(self, value: int) -> None:
        bytes_ = [(value >> (8 * i)) & 0xff for i in range(4)]
        self.append_bytes(bytes_)

    def write_big_integer(self, value: int) -> None:
        self.write_big_integer_string(str(value))

    def write_big_integer_string(self, value: str) -> None:
        if value == "0":
            bytes_ = [0]
        elif value.startswith("-"):
            raise ValueError("Unsigned bigint serialization not supported")
        else:
            hex_str = hex(int(value))[2:]
            if len(hex_str) % 2:
                hex_str = '0' + hex_str
            bytes_ = [int(hex_str[i:i+2], 16) for i in range(0, len(hex_str), 2)]
            bytes_.append(0)  # add sign at the end
        self.write_byte_array(bytes_)

    def write_signature(self, signature: Signature) -> None:
        if signature is None:
            self.write_byte(SignatureKind.None_.value)
        else:
            self.write_byte(signature.Kind.value)
            signature.SerializeData(self)

    def append_hex_encoded(self, hex_string: str) -> None:
        local_encoded = bytes.fromhex(hex_string)
        self.write_var_int(len(local_encoded))
        self.append_bytes(list(local_encoded))



# coding: utf-8

"""
    Phantasma API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class TransactionResult(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'hash': 'str',
        'chain_address': 'str',
        'timestamp': 'int',
        'block_height': 'int',
        'block_hash': 'str',
        'script': 'str',
        'payload': 'str',
        'events': 'list[EventResult]',
        'result': 'str',
        'fee': 'str',
        'state': 'str',
        'signatures': 'list[SignatureResult]',
        'sender': 'str',
        'gas_payer': 'str',
        'gas_target': 'str',
        'gas_price': 'str',
        'gas_limit': 'str',
        'expiration': 'int'
    }

    attribute_map = {
        'hash': 'hash',
        'chain_address': 'chainAddress',
        'timestamp': 'timestamp',
        'block_height': 'blockHeight',
        'block_hash': 'blockHash',
        'script': 'script',
        'payload': 'payload',
        'events': 'events',
        'result': 'result',
        'fee': 'fee',
        'state': 'state',
        'signatures': 'signatures',
        'sender': 'sender',
        'gas_payer': 'gasPayer',
        'gas_target': 'gasTarget',
        'gas_price': 'gasPrice',
        'gas_limit': 'gasLimit',
        'expiration': 'expiration'
    }

    def __init__(self, hash=None, chain_address=None, timestamp=None, block_height=None, block_hash=None, script=None, payload=None, events=None, result=None, fee=None, state=None, signatures=None, sender=None, gas_payer=None, gas_target=None, gas_price=None, gas_limit=None, expiration=None):  # noqa: E501
        """TransactionResult - a model defined in Swagger"""  # noqa: E501
        self._hash = None
        self._chain_address = None
        self._timestamp = None
        self._block_height = None
        self._block_hash = None
        self._script = None
        self._payload = None
        self._events = None
        self._result = None
        self._fee = None
        self._state = None
        self._signatures = None
        self._sender = None
        self._gas_payer = None
        self._gas_target = None
        self._gas_price = None
        self._gas_limit = None
        self._expiration = None
        self.discriminator = None
        if hash is not None:
            self.hash = hash
        if chain_address is not None:
            self.chain_address = chain_address
        if timestamp is not None:
            self.timestamp = timestamp
        if block_height is not None:
            self.block_height = block_height
        if block_hash is not None:
            self.block_hash = block_hash
        if script is not None:
            self.script = script
        if payload is not None:
            self.payload = payload
        if events is not None:
            self.events = events
        if result is not None:
            self.result = result
        if fee is not None:
            self.fee = fee
        if state is not None:
            self.state = state
        if signatures is not None:
            self.signatures = signatures
        if sender is not None:
            self.sender = sender
        if gas_payer is not None:
            self.gas_payer = gas_payer
        if gas_target is not None:
            self.gas_target = gas_target
        if gas_price is not None:
            self.gas_price = gas_price
        if gas_limit is not None:
            self.gas_limit = gas_limit
        if expiration is not None:
            self.expiration = expiration

    @property
    def hash(self):
        """Gets the hash of this TransactionResult.  # noqa: E501


        :return: The hash of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._hash

    @hash.setter
    def hash(self, hash):
        """Sets the hash of this TransactionResult.


        :param hash: The hash of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._hash = hash

    @property
    def chain_address(self):
        """Gets the chain_address of this TransactionResult.  # noqa: E501


        :return: The chain_address of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._chain_address

    @chain_address.setter
    def chain_address(self, chain_address):
        """Sets the chain_address of this TransactionResult.


        :param chain_address: The chain_address of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._chain_address = chain_address

    @property
    def timestamp(self):
        """Gets the timestamp of this TransactionResult.  # noqa: E501


        :return: The timestamp of this TransactionResult.  # noqa: E501
        :rtype: int
        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp):
        """Sets the timestamp of this TransactionResult.


        :param timestamp: The timestamp of this TransactionResult.  # noqa: E501
        :type: int
        """

        self._timestamp = timestamp

    @property
    def block_height(self):
        """Gets the block_height of this TransactionResult.  # noqa: E501


        :return: The block_height of this TransactionResult.  # noqa: E501
        :rtype: int
        """
        return self._block_height

    @block_height.setter
    def block_height(self, block_height):
        """Sets the block_height of this TransactionResult.


        :param block_height: The block_height of this TransactionResult.  # noqa: E501
        :type: int
        """

        self._block_height = block_height

    @property
    def block_hash(self):
        """Gets the block_hash of this TransactionResult.  # noqa: E501


        :return: The block_hash of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._block_hash

    @block_hash.setter
    def block_hash(self, block_hash):
        """Sets the block_hash of this TransactionResult.


        :param block_hash: The block_hash of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._block_hash = block_hash

    @property
    def script(self):
        """Gets the script of this TransactionResult.  # noqa: E501


        :return: The script of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._script

    @script.setter
    def script(self, script):
        """Sets the script of this TransactionResult.


        :param script: The script of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._script = script

    @property
    def payload(self):
        """Gets the payload of this TransactionResult.  # noqa: E501


        :return: The payload of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._payload

    @payload.setter
    def payload(self, payload):
        """Sets the payload of this TransactionResult.


        :param payload: The payload of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._payload = payload

    @property
    def events(self):
        """Gets the events of this TransactionResult.  # noqa: E501


        :return: The events of this TransactionResult.  # noqa: E501
        :rtype: list[EventResult]
        """
        return self._events

    @events.setter
    def events(self, events):
        """Sets the events of this TransactionResult.


        :param events: The events of this TransactionResult.  # noqa: E501
        :type: list[EventResult]
        """

        self._events = events

    @property
    def result(self):
        """Gets the result of this TransactionResult.  # noqa: E501


        :return: The result of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._result

    @result.setter
    def result(self, result):
        """Sets the result of this TransactionResult.


        :param result: The result of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._result = result

    @property
    def fee(self):
        """Gets the fee of this TransactionResult.  # noqa: E501


        :return: The fee of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._fee

    @fee.setter
    def fee(self, fee):
        """Sets the fee of this TransactionResult.


        :param fee: The fee of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._fee = fee

    @property
    def state(self):
        """Gets the state of this TransactionResult.  # noqa: E501


        :return: The state of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._state

    @state.setter
    def state(self, state):
        """Sets the state of this TransactionResult.


        :param state: The state of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._state = state

    @property
    def signatures(self):
        """Gets the signatures of this TransactionResult.  # noqa: E501


        :return: The signatures of this TransactionResult.  # noqa: E501
        :rtype: list[SignatureResult]
        """
        return self._signatures

    @signatures.setter
    def signatures(self, signatures):
        """Sets the signatures of this TransactionResult.


        :param signatures: The signatures of this TransactionResult.  # noqa: E501
        :type: list[SignatureResult]
        """

        self._signatures = signatures

    @property
    def sender(self):
        """Gets the sender of this TransactionResult.  # noqa: E501


        :return: The sender of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._sender

    @sender.setter
    def sender(self, sender):
        """Sets the sender of this TransactionResult.


        :param sender: The sender of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._sender = sender

    @property
    def gas_payer(self):
        """Gets the gas_payer of this TransactionResult.  # noqa: E501


        :return: The gas_payer of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._gas_payer

    @gas_payer.setter
    def gas_payer(self, gas_payer):
        """Sets the gas_payer of this TransactionResult.


        :param gas_payer: The gas_payer of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._gas_payer = gas_payer

    @property
    def gas_target(self):
        """Gets the gas_target of this TransactionResult.  # noqa: E501


        :return: The gas_target of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._gas_target

    @gas_target.setter
    def gas_target(self, gas_target):
        """Sets the gas_target of this TransactionResult.


        :param gas_target: The gas_target of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._gas_target = gas_target

    @property
    def gas_price(self):
        """Gets the gas_price of this TransactionResult.  # noqa: E501


        :return: The gas_price of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._gas_price

    @gas_price.setter
    def gas_price(self, gas_price):
        """Sets the gas_price of this TransactionResult.


        :param gas_price: The gas_price of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._gas_price = gas_price

    @property
    def gas_limit(self):
        """Gets the gas_limit of this TransactionResult.  # noqa: E501


        :return: The gas_limit of this TransactionResult.  # noqa: E501
        :rtype: str
        """
        return self._gas_limit

    @gas_limit.setter
    def gas_limit(self, gas_limit):
        """Sets the gas_limit of this TransactionResult.


        :param gas_limit: The gas_limit of this TransactionResult.  # noqa: E501
        :type: str
        """

        self._gas_limit = gas_limit

    @property
    def expiration(self):
        """Gets the expiration of this TransactionResult.  # noqa: E501


        :return: The expiration of this TransactionResult.  # noqa: E501
        :rtype: int
        """
        return self._expiration

    @expiration.setter
    def expiration(self, expiration):
        """Sets the expiration of this TransactionResult.


        :param expiration: The expiration of this TransactionResult.  # noqa: E501
        :type: int
        """

        self._expiration = expiration

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(TransactionResult, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, TransactionResult):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



import re


class EventDecoder():
    '''
    The EventDecoder object implements all the basic methods needed to decode the data related to 
    Phantasma blockchain events.

    Args:
        data (str): encoded event data.

    Attributes:
        data (str): encoded event data.
    '''

    data: str = None

    def __init__(self, data: str) -> None:
        self.data = data

    def readCharPair(self) -> str:
        '''This method reads and returns a pair of characters from the encoded data and updates the object data
        with the remainder characters.

        Args:
            None

        Returns:
            readCharPair (str): char pair data read
        '''
        res = self.data[0:2]
        self.data = self.data[2:]
        return res

    def readByte(self) -> int:
        '''This method reads a byte from the encoded data.

        Args:
            None

        Returns:
            readByte (int): byte data read.
        '''
        b = self.readCharPair()
        return int(b, 16)

    def read(self, numBytes: int) -> str:
        '''This method reads and returns the number of numBytes from encoded data and updates the object data
        with the remainder characters

        Args:
            numBytes (int): number of bytes to be read.

        Returns:
            read (int): string data read.
        '''

        b = (numBytes * 2)
        res = self.data[0:b]
        self.data = self.data[b:]
        return res

    def readString(self) -> str:
        '''This method reads a str value from encoded data.

        Args:
            None

        Returns:
            readString (str): string data read.
        '''

        le = self.readVarInt()
        return self.readStringBytes(le)

    def readStringBytes(self, numBytes: int) -> str:
        '''This method reads a str bytes from encoded data.

        Args:
            numBytes (int): number of bytes to be read.

        Returns:
            readString (str): string data read.
        '''

        res = ""
        for i in range(numBytes):
            res = res + chr(self.readByte())
        return res

    def readVarInt(self) -> int:
        '''This method reads int value from the encoded data.

        Args:
            None

        Returns:
            readVarInt (int): int data read.
        '''

        le = self.readByte()
        res = 0
        if (le == 0xfd):
            d = re.findall('.{1,2}', self.read(2))
            for c in reversed(d):
                res = (res * 256 + int(c, 16))
            return res
        if (le == 0xfe):
            d = re.findall('.{1,2}', self.read(4))
            for c in reversed(d):
                res = (res * 256 + int(c, 16))
            return res
        if (le == 0xff):
            d = re.findall('.{1,2}', self.read(8))
            for c in reversed(d):
                res = (res * 256 + int(c, 16))
            return res
        return le

    def readBigIntAccurate(self) -> int:
        '''This method reads int value from the encoded data.

        Args:
            None

        Returns:
            readVarInt (int): int data read.
        '''
        le = self.readVarInt()
        res = 0
        stringBytes = self.read(le)
        d = re.findall('.{1,2}', stringBytes)
        for c in reversed(d):
            b = int(c, 16)
            res = (res * 256 + b)
        return res



#from Types import PhantasmaKeys
from Types.Enums import AddressKind
from Types import Serialization
import base58
import hashlib
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519
from typing import Union

class Address:
    NullText = "NULL"
    LengthInBytes = 34
    MaxPlatformNameLength = 10
    NullPublicKey = bytes([0] * LengthInBytes)
    Null = None  # This will be initialized after the class definition

    _keyToTextCache = {}

    def __init__(self, public_key: bytes):
        if len(public_key) != Address.LengthInBytes:
            raise ValueError(f"publicKey length must be {Address.LengthInBytes}, it was {len(public_key)}")
        self._bytes = public_key
        self._text = None

    @staticmethod
    def FromKey(key):
        bytes = bytearray()
        bytes.extend([ AddressKind.User])
        if len(key) == 32:
            bytes.extend([0] * 1) 
            bytes.extend(key)
        elif len(key) == 33:
            #bytes.extend([0] * 1) 
            bytes.extend(key[1:])
        elif len(key) == 64:
            #bytes.extend([0] * 1) 
            bytes.extend(key[:32])
        else:
            raise ValueError(f'Invalid public key length: {len(key)}')
        return Address(bytes)

    def Kind(self):
        if self.IsNull():
            return AddressKind.System
        elif self._bytes[0] >= 3:
            return AddressKind.Interop
        else:
            return chr(self._bytes[0])

    def IsSystem(self) -> bool:
        return self.Kind == AddressKind.System

    def IsInterop(self) -> bool:
        return self.Kind == AddressKind.Interop

    def IsUser(self) -> bool:
        return self.Kind == AddressKind.User
    
    def IsNull(self):
        if self._bytes is None or len(self._bytes) == 0:
            return True

        for i in range(1, len(self._bytes)):
            if self._bytes[i] != 0:
                return False

        return True

    @property
    def Text(self):
        if self.IsNull():
            return Address.NullText

        if not self._text:
            #if self._bytes in self._keyToTextCache:
            #    print("here3")
            #    self._text = self._keyToTextCache[self._bytes]

            if not self._text:
                prefix = 'P' 
                if self.IsNull():
                    prefix = 'S'
                elif self._bytes[0] >= 3:
                    prefix = 'X'
                base58_text = base58.b58encode(self._bytes).decode('utf-8')
                self._text = prefix + base58_text
                #self._keyToTextCache[self._bytes] = self._text

        return self._text
    
    @staticmethod
    def FromPublicKey(public_key: bytes):
        public_key = public_key[:Address.LengthInBytes]
        return Address(public_key)

    @staticmethod
    def FromText(text: str):
        return Address.Parse(text)

    @staticmethod
    def Parse(text: str):
        if text is None:
            return Address.Null

        if text == Address.NullText:
            return Address.Null

        prefix = text[0]
        text = text[1:]
        bytes_ = base58.b58decode(text)
        addr = Address(bytes_)

        if prefix == "P" and addr.Kind != AddressKind.User:
            raise ValueError("Invalid address prefix. Expected 'P'")
        elif prefix == "S" and addr.Kind != AddressKind.System:
            raise ValueError("Invalid address prefix. Expected 'S'")
        elif prefix == "X" and addr.Kind < AddressKind.Interop:
            raise ValueError("Invalid address prefix. Expected 'X'")

        return addr

    @staticmethod
    def IsValidAddress(text: str) -> bool:
        try:
            Address.FromText(text)
            return True
        except:
            return False

    @staticmethod
    def FromHash(input):
        if isinstance(input, str):
            bytes_input = input.encode()
        else:
            bytes_input = input

        hash = hashlib.sha256(bytes_input).digest()
        address_bytes = bytearray(Address.LengthInBytes)
        address_bytes[0] = AddressKind.User
        address_bytes[2:34] = hash[:32]
        return Address(bytes(address_bytes))

    @staticmethod
    def FromWif(wif):
        # Assuming `get_private_key_from_wif` function is implemented to extract the private key from WIF
        private_key = get_private_key_from_wif(wif)
        public_key = private_key.public_key().public_bytes(
            encoding=Serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw
        )
        address_hex = b'\x01\x00' + public_key
        return Address.FromBytes(address_hex)

    @staticmethod
    def FromBytes(bytes):
        return Address(bytes)

    def compareTo(self, other):
        return (self._bytes > other._bytes) - (self._bytes < other._bytes)

    def equals(self, other):
        return isinstance(other, Address) and self._bytes == other._bytes

    def __eq__(self, other):
        return self.equals(other)

    def __str__(self):
        if self.is_null():
            return self.NullText

        prefix = {AddressKind.User: 'P', AddressKind.Interop: 'X'}.get(self.kind(), 'S')
        return prefix + base58.b58encode(self._bytes).decode()

    def to_byte_array(self):
        return self._bytes

    def serialize_data(self, writer):
        # Assuming PBinaryWriter is implemented
        writer.write_byte_array(self._bytes)

    def unserialize_data(self, reader):
        # Assuming PBinaryReader is implemented
        self._bytes = reader.read_byte_array()
        self._text = None

    def is_null(self):
        return self._bytes == bytearray(self.LengthInBytes)

    def kind(self):
        """
        Determine the address kind based on the first byte.
        This logic should be aligned with how your application defines different address kinds.
        """
        if self._bytes[0] == AddressKind.User:
            return "User"
        elif self._bytes[0] == AddressKind.Interop:
            return "Interop"
        # Add more conditions for other address kinds as needed
        else:
            return "Unknown"


from datetime import datetime
from typing import Any
from enum import Enum, IntEnum
from .Opcode import Opcode
from .VMType import VMType


class ScriptBuilder():

    '''
    The ScriptBuilder object implements all the methods needed to build a Phantasma Script string
    which is needed to interact with the Phantasma Blockchain.

    Args:
        None

    Attributes:
        data (str): This is where we store the script string.
        Nexus (dict): Dictionary which contains Phantasma Blockchain contract names.
        nullAddress (str): Phantasma Blockchain NULL address representation.
        _labelLocations (list): Label's location list
        _jumpLocations (list): Jump's location list
    '''

    Nexus = dict(
        GasContractName='gas',
        BlockContractName='block',
        StakeContractName='stake',
        SwapContractName='swap',
        AccountContractName='account',
        ConsensusContractName='consensus',
        GovernanceContractName='governance',
        StorageContractName='storage',
        ValidatorContractName='validator',
        InteropContractName='interop',
        ExchangeContractName='exchange',
        PrivacyContractName='privacy',
        RelayContractName='relay',
        RankingContractName='ranking'
    )

    MAX_REGISTER_COUNT = 32

    data: str = None
    _labelLocations = []
    _jumpLocations = []

    nullAddress = "S1111111111111111111111111111111111"

    def __init__(self) -> None:
        self.data = ""
        self._labelLocations = []
        self._jumpLocations = []

    def BeginScript(self) -> None:
        '''This method initializes an empty script.

        Args:
            None

        Returns:
            None
        '''
        
        self.data = ""
        self._labelLocations = []
        self._jumpLocations = []

    def GetScript(self) -> str:
        '''This method returns the generated script as string.

        Args:
            None

        Returns:
            getScript (str): end result script.
        '''
        return self.data

    def EndScript(self) -> str:
        '''This method finishes the script and return it as string.

        Args:
            None

        Returns:
            endScript (str): end result script.
        '''
        self.Emit(Opcode.RET)
        return self.data

    def Emit(self, opcode: Opcode, data: Any = None) -> None:
        '''This method emits an opcode with its data.

        Args:
            opcode (Opcode): Blockchain opcode to be emitted.
            data (Any): Data related to the opcode to be emitted.

        Returns:
            None
        '''
        self.AppendByte(opcode)
        if data is not None:
            if isinstance(data, list):
                for d in data:
                    self.emit(d)
            else:
                self.emit(data)
        return self

    def EmitPush(self, reg: int) -> None:
        '''This method emits a PUSH opcode.

        Args:
            reg (int): register

        Returns:
            None
        '''
        self.Emit(Opcode.PUSH)
        self.AppendByte(reg)
        return self

    def EmitPop(self, reg: int) -> None:
        '''This method emits a POP opcode.

        Args:
            reg (int): register

        Returns:
            None
        '''
        self.Emit(Opcode.POP)
        self.AppendByte(reg)
        return self

    def EmitExtCall(self, method: str, reg: int = 0) -> None:
        '''This method emits a EXTCALL opcode.

        Args:
            method (str): method to be executed.
            reg (int): register

        Returns:
            None
        '''
        self.EmitLoad(reg, method)
        self.Emit(Opcode.EXTCALL)
        self.AppendByte(reg)
        return self

    def RawString(self, value: str) -> list:
        '''This method converts an string to a list of byte characters.

        Args:
            value (str): string to be converted.

        Returns:
            rawString (list): list of byte chars.
        '''
        data = []
        for c in list(value):
            data.append(ord(c))
        return data

    def EmitLoad(self, reg: int, obj: Any) -> None:
        '''This method loads data to the script depending of the object type
        Supported object types are: str, bool, int, float and datetime.

        Args:
            reg (int): register
            obj (Any): data

        Returns:
            None
        '''

        # String
        typeLoaded = False
        if (isinstance(obj, str)):
            data = self.rawString(obj)
            self.emitLoadBytes(reg, data, VMType.String)
            typeLoaded = True

        # Boolean
        if (isinstance(obj, bool)):
            data = []
            if obj:
                data.append(1)
            else:
                data.append(0)
            self.emitLoadBytes(reg, data, VMType.Bool)
            typeLoaded = True

        # Number
        if ((isinstance(obj, int)) or (isinstance(obj, float))):
            data = self.rawString(str(obj))
            self.emitLoadBytes(reg, data, VMType.String)
            typeLoaded = True

        # Timestamp
        if (isinstance(obj, datetime)):
            data = self.rawString(str(obj))
            self.emitLoadTimestamp(reg, obj)
            typeLoaded = True

        if not typeLoaded:
            raise Exception("Load type " + str(type(obj)) +
                            " is not supported")

        return self

    def EmitLoadBytes(self, reg: int, data: bytes = None,
                      typ: VMType = VMType.Bytes) -> None:

        '''This method loads bytes data to the script.

        Args:
            reg (int): register.
            data (bytes): data.
            typ (VMType): data type.

        Returns:
            None
        '''

        if data is None:
            data = []

        if len(data) > 0xffff:
            raise Exception("tried to load too much data")

        self.Emit(Opcode.LOAD)
        self.AppendByte(reg)
        self.AppendByte(typ)
        self.EmitVarInt(len(data))
        self.EmitBytes(data)
        return self

    def EmitLoadEnum(self, reg: int, val: int) -> None:
        '''This method loads enum data to the script.

        Args:
            reg (int): register.
            val (int): data.

        Returns:
            None
        '''

        data = [0, 0, 0, 0]

        for i in range(len(data)):
            data[c] = (val & 0xff)
            val = (val - (val & 0xff)) / 256

        self.EmitLoadBytes(reg, data, VMType.Enum)
        return self

    def EmitLoadTimestamp(self, reg: int, obj: datetime) -> None:
        '''This method loads a datetime data to the script as an UTC timestamp.

        Args:
            reg (int): register.
            obj (datetime): data.

        Returns:
            None
        '''

        num = int(obj.replace(tzinfo=timezone.utc).timestamp())

        a = (num & 0xff000000) >> 24
        b = (num & 0x00ff0000) >> 16
        c = (num & 0x0000ff00) >> 8
        d = num & 0x000000ff

        data = [d, c, b, a]
        self.EmitLoadBytes(reg, data, VMType.Timestamp)
        return self

    def EmitMove(self, src_reg: int, dst_reg: int) -> None:
        '''This method emits a MOVE opcode.

        Args:
            src_reg (int): source register.
            dst_reg (int): destination register.

        Returns:
            None
        '''

        self.Emit(Opcode.MOVE)
        self.AppendByte(src_reg)
        self.AppendByte(dst_reg)
        return self

    def EmitCopy(self, src_reg: int, dst_reg: int) -> None:
        '''This method emits a COPY opcode.

        Args:
            src_reg (int): source register.
            dst_reg (int): destination register.

        Returns:
            None
        '''

        self.Emit(Opcode.COPY)
        self.AppendByte(src_reg)
        self.AppendByte(dst_reg)
        return self

    def EmitLabel(self, label: str) -> None:
        '''This method loads a label into the script.

        Args:
            label (str): label data.

        Returns:
            None
        '''

        self.Emit(Opcode.NOP)
        self._labelLocations[label] = len(self.data)
        return self

    def EmitJump(self, opcode: Opcode, label: str, reg: int = 0) -> None:
        '''This method emits a JUMP opcode.

        Args:
            opcode (Opcode): jump opcode to add: Opcode.JMP, Opcode.JMPIF or Opcode.JMPNOT
            label (str): label to jump.
            reg (int): register.

        Returns:
            None
        '''

        if ((opcode == Opcode.JMP) or (
                opcode == Opcode.JMPIF) or (opcode == Opcode.JMPNOT)):
            self.Emit(opcode)
        else:
            raise Exception("Invalid jump opcode: " + str(opcode))

        if (opcode != Opcode.JMP):
            self.AppendByte(reg)

        ofs = len(self.data)
        self.AppendUshort(0)
        self._jumpLocations[ofs] = label
        return self

    def EmitCall(self, label: str, regCount: int) -> None:
        '''This method emits a CALL opcode.

        Args:
            label (str): label to call.
            regCount (int): register counter.

        Returns:
            None
        '''

        if ((regCount < 1) or (regCount > MAX_REGISTER_COUNT)):
            raise Exception("Invalid number of registers")

        ofs = (len(self.data)) + 2
        self.Emit(Opcode.CALL)
        self.AppendUshort(0)
        self._jumpLocations[ofs] = label
        return self

    def EmitConditionalJump(
            self,
            opcode: Opcode,
            src_reg: int,
            label: str) -> None:
        '''This method emits a conditional JUMP opcode.

        Args:
            opcode (Opcode): jump opcode to add: only Opcode.JMPIF and Opcode.JMPNOT are valid.
            label (str): label to jump.
            src_reg (int): source register.

        Returns:
            None
        '''

        if ((opcode != Opcode.JMPIF) and (
                opcode != Opcode.JMPNOT)):
            raise Exception("Opcode is not a conditional jump: " + str(opcode))

        ofs = (len(self.data)) + 2
        self.Emit(opcode)
        self.AppendUshort(0)
        self._jumpLocations[ofs] = label
        return self

    def InsertMethodArgs(self, args: list) -> None:
        '''This loads method arguments into the script.

        Args:
            args (list): list of arguments.

        Returns:
            None
        '''

        temp_reg = 0
        for arg in reversed(args):
            self.EmitLoad(temp_reg, arg)
            self.EmitPush(temp_reg)

    def CallInterop(self, method: str, args: list) -> None:
        '''This method loads an external method and its arguments into the script.

        Args:
            method (str): method to be executed.
            args (list): list of arguments.

        Returns:
            None
        '''

        self.insertMethodArgs(args)

        dest_reg = 0
        self.emitLoad(dest_reg, method)
        self.emit(Opcode.EXTCALL, dest_reg)
        return self

    def CallContract(self, contractName: str, method: str, args: list) -> None:
        '''This method adds a contract method call with its arguments into the script.

        Args:
            contractName (str): contract name to be executed.
            method (str): method to be executed.
            args (list): list of arguments.

        Returns:
            None
        '''

        self.insertMethodArgs(args)

        temp_reg = 0
        self.emitLoad(temp_reg, method)
        self.emitPush(temp_reg)

        src_reg = 0
        dest_reg = 1
        self.emitLoad(src_reg, contractName)
        self.emit(Opcode.CTX, [src_reg, dest_reg])
        self.emit(Opcode.SWITCH, [dest_reg])
        return self

    def EmitVarString(self, text: str) -> None:
        '''This method loads a str variable into the script.

        Args:
            text (str): string data.

        Returns:
            None
        '''

        data = self.RawString(text)
        self.EmitVarInt(len(data))
        self.EmitBytes(data)
        return self

    def EmitVarInt(self, value: int) -> None:
        '''This method loads a int variable into the script.

        Args:
            value (int): int data.

        Returns:
            None
        '''

        if value < 0:
            raise Exception("Negative value invalid")
        if (value < 0xfd):
            self.appendByte(value)
        elif (value <= 0xffff):
            B = (value & 0x0000ff00) >> 8
            A = value & 0x000000ff
            self.appendByte(0xfd)
            self.appendByte(A)
            self.appendByte(B)
        elif (value <= 0xffffffff):
            C = (value & 0x00ff0000) >> 16
            B = (value & 0x0000ff00) >> 8
            A = value & 0x000000ff
            self.appendByte(0xfe)
            self.appendByte(A)
            self.appendByte(B)
            self.appendByte(C)
        else:
            D = (value & 0xff000000) >> 24
            C = (value & 0x00ff0000) >> 16
            B = (value & 0x0000ff00) >> 8
            A = value & 0x000000ff
            self.appendByte(0xff)
            self.appendByte(A)
            self.appendByte(B)
            self.appendByte(C)
            self.appendByte(D)
        return self

    def EmitBytes(self, data: bytes) -> None:
        '''This method loads a bytes variable into the script.

        Args:
            data (bytes): bytes data.

        Returns:
            None
        '''

        for i in data:
            self.appendByte(i)

    def ByteToHex(self, data: int) -> str:
        '''This method converts a byte into an str hex representation.

        Args:
            data (int): byte data to be converted to hex.

        Returns:
            byteToHex (str): str hex representatiom of the byte.
        '''

        res = format(data, 'x').upper()
        if len(res) == 1:
            res = "0" + res
        return res

    def AppendByte(self, data: int) -> None:
        '''This method loads a byte variable into the script.

        Args:
            data (int): int data.

        Returns:
            None
        '''

        self.data = self.data + self.byteToHex(data)

    def AppendUshort(self, ushort: int) -> None:
        '''This method loads a unsigned short int variable into the script.

        Args:
            data (int): unsigned short int data.

        Returns:
            None
        '''

        self.data = self.data + \
            (self.byteToHex(ushort & 0xff)) + (this.byteToHex((ushort >> 8) & 0xff))

    def AppendHexEncoded(self, data: str) -> None:
        '''This method loads an Hex string variable into the script.

        Args:
            data (str): Hex data to be loaded.

        Returns:
            None
        '''

        self.data = self.data + data
        return self

    # ScriptBuilderExtensions
    def AllowGas(
            self,
            frm: str,
            to: str,
            gasPrice: int,
            gasLimit: int) -> None:
        return self.callContract(
            self.Nexus["GasContractName"], "AllowGas", [
                frm, to, gasPrice, gasLimit])
        '''This method is a wrapper that invokes the method AllowGas from the gas contract.

        Args:
            frm (str): Phantasma wallet address where gas fees are going to be spent.
            to (str): Phantasma wallet address where gas fees are going to be sent (NULL Address).
            gasPrice (int): Max gas price to used.
            gasLimit (int): Gas limit to be used.

        Returns:
            None
        '''

    def SpendGas(self, address: str) -> None:
        return self.callContract(
            self.Nexus["GasContractName"], "SpendGas", [address])
        '''This method is a wrapper that invokes the method SpendGas from the gas contract.

        Args:
            address (str): Phantasma wallet address where gas fees are going to be spent.

        Returns:
            None
        '''
           



# coding: utf-8

"""
    Phantasma API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class ChainResult(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'name': 'str',
        'address': 'str',
        'parent': 'str',
        'height': 'int',
        'organization': 'str',
        'contracts': 'list[str]',
        'dapps': 'list[str]'
    }

    attribute_map = {
        'name': 'name',
        'address': 'address',
        'parent': 'parent',
        'height': 'height',
        'organization': 'organization',
        'contracts': 'contracts',
        'dapps': 'dapps'
    }

    def __init__(self, name=None, address=None, parent=None, height=None, organization=None, contracts=None, dapps=None):  # noqa: E501
        """ChainResult - a model defined in Swagger"""  # noqa: E501
        self._name = None
        self._address = None
        self._parent = None
        self._height = None
        self._organization = None
        self._contracts = None
        self._dapps = None
        self.discriminator = None
        if name is not None:
            self.name = name
        if address is not None:
            self.address = address
        if parent is not None:
            self.parent = parent
        if height is not None:
            self.height = height
        if organization is not None:
            self.organization = organization
        if contracts is not None:
            self.contracts = contracts
        if dapps is not None:
            self.dapps = dapps

    @property
    def name(self):
        """Gets the name of this ChainResult.  # noqa: E501


        :return: The name of this ChainResult.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ChainResult.


        :param name: The name of this ChainResult.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def address(self):
        """Gets the address of this ChainResult.  # noqa: E501


        :return: The address of this ChainResult.  # noqa: E501
        :rtype: str
        """
        return self._address

    @address.setter
    def address(self, address):
        """Sets the address of this ChainResult.


        :param address: The address of this ChainResult.  # noqa: E501
        :type: str
        """

        self._address = address

    @property
    def parent(self):
        """Gets the parent of this ChainResult.  # noqa: E501


        :return: The parent of this ChainResult.  # noqa: E501
        :rtype: str
        """
        return self._parent

    @parent.setter
    def parent(self, parent):
        """Sets the parent of this ChainResult.


        :param parent: The parent of this ChainResult.  # noqa: E501
        :type: str
        """

        self._parent = parent

    @property
    def height(self):
        """Gets the height of this ChainResult.  # noqa: E501


        :return: The height of this ChainResult.  # noqa: E501
        :rtype: int
        """
        return self._height

    @height.setter
    def height(self, height):
        """Sets the height of this ChainResult.


        :param height: The height of this ChainResult.  # noqa: E501
        :type: int
        """

        self._height = height

    @property
    def organization(self):
        """Gets the organization of this ChainResult.  # noqa: E501


        :return: The organization of this ChainResult.  # noqa: E501
        :rtype: str
        """
        return self._organization

    @organization.setter
    def organization(self, organization):
        """Sets the organization of this ChainResult.


        :param organization: The organization of this ChainResult.  # noqa: E501
        :type: str
        """

        self._organization = organization

    @property
    def contracts(self):
        """Gets the contracts of this ChainResult.  # noqa: E501


        :return: The contracts of this ChainResult.  # noqa: E501
        :rtype: list[str]
        """
        return self._contracts

    @contracts.setter
    def contracts(self, contracts):
        """Sets the contracts of this ChainResult.


        :param contracts: The contracts of this ChainResult.  # noqa: E501
        :type: list[str]
        """

        self._contracts = contracts

    @property
    def dapps(self):
        """Gets the dapps of this ChainResult.  # noqa: E501


        :return: The dapps of this ChainResult.  # noqa: E501
        :rtype: list[str]
        """
        return self._dapps

    @dapps.setter
    def dapps(self, dapps):
        """Sets the dapps of this ChainResult.


        :param dapps: The dapps of this ChainResult.  # noqa: E501
        :type: list[str]
        """

        self._dapps = dapps

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(ChainResult, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ChainResult):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    Phantasma API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class ResponseQuery(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'code': 'int',
        'log': 'str',
        'info': 'str',
        'index': 'str',
        'key': 'str',
        'value': 'str',
        'proof': 'str',
        'height': 'str',
        'codespace': 'str'
    }

    attribute_map = {
        'code': 'code',
        'log': 'log',
        'info': 'info',
        'index': 'index',
        'key': 'key',
        'value': 'value',
        'proof': 'proof',
        'height': 'height',
        'codespace': 'codespace'
    }

    def __init__(self, code=None, log=None, info=None, index=None, key=None, value=None, proof=None, height=None, codespace=None):  # noqa: E501
        """ResponseQuery - a model defined in Swagger"""  # noqa: E501
        self._code = None
        self._log = None
        self._info = None
        self._index = None
        self._key = None
        self._value = None
        self._proof = None
        self._height = None
        self._codespace = None
        self.discriminator = None
        if code is not None:
            self.code = code
        if log is not None:
            self.log = log
        if info is not None:
            self.info = info
        if index is not None:
            self.index = index
        if key is not None:
            self.key = key
        if value is not None:
            self.value = value
        if proof is not None:
            self.proof = proof
        if height is not None:
            self.height = height
        if codespace is not None:
            self.codespace = codespace

    @property
    def code(self):
        """Gets the code of this ResponseQuery.  # noqa: E501


        :return: The code of this ResponseQuery.  # noqa: E501
        :rtype: int
        """
        return self._code

    @code.setter
    def code(self, code):
        """Sets the code of this ResponseQuery.


        :param code: The code of this ResponseQuery.  # noqa: E501
        :type: int
        """

        self._code = code

    @property
    def log(self):
        """Gets the log of this ResponseQuery.  # noqa: E501


        :return: The log of this ResponseQuery.  # noqa: E501
        :rtype: str
        """
        return self._log

    @log.setter
    def log(self, log):
        """Sets the log of this ResponseQuery.


        :param log: The log of this ResponseQuery.  # noqa: E501
        :type: str
        """

        self._log = log

    @property
    def info(self):
        """Gets the info of this ResponseQuery.  # noqa: E501


        :return: The info of this ResponseQuery.  # noqa: E501
        :rtype: str
        """
        return self._info

    @info.setter
    def info(self, info):
        """Sets the info of this ResponseQuery.


        :param info: The info of this ResponseQuery.  # noqa: E501
        :type: str
        """

        self._info = info

    @property
    def index(self):
        """Gets the index of this ResponseQuery.  # noqa: E501


        :return: The index of this ResponseQuery.  # noqa: E501
        :rtype: str
        """
        return self._index

    @index.setter
    def index(self, index):
        """Sets the index of this ResponseQuery.


        :param index: The index of this ResponseQuery.  # noqa: E501
        :type: str
        """

        self._index = index

    @property
    def key(self):
        """Gets the key of this ResponseQuery.  # noqa: E501


        :return: The key of this ResponseQuery.  # noqa: E501
        :rtype: str
        """
        return self._key

    @key.setter
    def key(self, key):
        """Sets the key of this ResponseQuery.


        :param key: The key of this ResponseQuery.  # noqa: E501
        :type: str
        """

        self._key = key

    @property
    def value(self):
        """Gets the value of this ResponseQuery.  # noqa: E501


        :return: The value of this ResponseQuery.  # noqa: E501
        :rtype: str
        """
        return self._value

    @value.setter
    def value(self, value):
        """Sets the value of this ResponseQuery.


        :param value: The value of this ResponseQuery.  # noqa: E501
        :type: str
        """

        self._value = value

    @property
    def proof(self):
        """Gets the proof of this ResponseQuery.  # noqa: E501


        :return: The proof of this ResponseQuery.  # noqa: E501
        :rtype: str
        """
        return self._proof

    @proof.setter
    def proof(self, proof):
        """Sets the proof of this ResponseQuery.


        :param proof: The proof of this ResponseQuery.  # noqa: E501
        :type: str
        """

        self._proof = proof

    @property
    def height(self):
        """Gets the height of this ResponseQuery.  # noqa: E501


        :return: The height of this ResponseQuery.  # noqa: E501
        :rtype: str
        """
        return self._height

    @height.setter
    def height(self, height):
        """Sets the height of this ResponseQuery.


        :param height: The height of this ResponseQuery.  # noqa: E501
        :type: str
        """

        self._height = height

    @property
    def codespace(self):
        """Gets the codespace of this ResponseQuery.  # noqa: E501


        :return: The codespace of this ResponseQuery.  # noqa: E501
        :rtype: str
        """
        return self._codespace

    @codespace.setter
    def codespace(self, codespace):
        """Sets the codespace of this ResponseQuery.


        :param codespace: The codespace of this ResponseQuery.  # noqa: E501
        :type: str
        """

        self._codespace = codespace

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(ResponseQuery, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseQuery):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other
